# -*- coding: utf-8 -*-
"""
Dynamic programming algorithm to solve nonograms (using recursion)

See details in the work 'An Efficient Approach to Solving Nonograms':
https://ir.nctu.edu.tw/bitstream/11536/22772/1/000324586300005.pdf
"""

from six.moves import zip

from ..core.common import BOX, SPACE, UNKNOWN, partial_sums
from .base import BaseLineSolver, NonogramError


class EfficientSolver(BaseLineSolver):
    """
    Recursive nonogram line solver.
    Adapted from the work 'An Efficient Approach to Solving Nonograms'
    """

    def __init__(self, description, line):
        super(EfficientSolver, self).__init__(description, line)

        self.minimum_lengths = self.min_lengths(self.description)
        self.additional_space = self._set_additional_space()

        self._cache_width = len(self.description) + 1
        self._fix_table = self._init_tables()
        self._paint_table = self._init_tables()

    def _init_tables(self):
        cache_height = len(self.line) + 1
        return [None] * (self._cache_width * cache_height)

    def _linear_index(self, i, j):
        return (i + 1) * self._cache_width + (j + 1)

    def _set_additional_space(self):
        space = self.empty_cell()
        if self.line[0] != space:
            self.line = (space,) + self.line
            return True
        return False

    @classmethod
    def min_lengths(cls, description):
        """
        The minimum line sizes in which can squeeze from 0 to i-th block
        """

        min_indexes = [s - 1 for s in partial_sums(description, colored=False)]
        return min_indexes

    def fix(self, i, j):
        """
        Verify whether blocks from 0 to (j-1)-th
        can be resided inside a substring line[:i+1]
        """

        fixable = self._fix_table[self._linear_index(i, j)]
        if fixable is None:
            fixable = self._fix(i, j)
            self._fix_table[self._linear_index(i, j)] = fixable
        return fixable

    @classmethod
    def _can_be_space(cls, cell):
        return cell in (SPACE, UNKNOWN)

    def _fix_border_conditions(self, i, j):
        if j < 0:
            assert j == -1

            if i < 0:
                return True

            # NB: improvement
            return all(map(self._can_be_space, self.line[:i + 1]))

        # reached the beginning of the line
        if i < 0:
            assert i == -1

            # no more blocks to fill
            return j < 0

        if i < self.minimum_lengths[j]:
            return False

        return None

    def _fix(self, i, j):
        """
        Determine whether S[:i+1] is fixable with respect to D[:j+1]
        :param i: line size
        :param j: block number
        """
        res = self._fix_border_conditions(i, j)
        if res is not None:
            return res

        res = self._fix0(i, j) or self._fix1(i, j)
        return res

    def _fix0(self, i, j):
        """
        Determine whether S[:i+1] is fixable with respect to D[:j+1]
        in assumption that S[i] can be 0
        :param i: line size
        :param j: block number
        """

        if self._can_be_space(self.line[i]):
            return self.fix(i - 1, j)

        return False

    def _fix1(self, i, j):
        """
        Determine whether S[:i+1] is fixable with respect to D[:j+1]
        in assumption that S[i] can be 1
        :param i: line size
        :param j: block number
        """
        block_size = self.description[j]
        if j >= 0 and i >= block_size:
            block = self.line[i - block_size: i + 1]
            if self._is_space_with_block(block):
                return self.fix(i - block_size - 1, j - 1)

        return False

    @classmethod
    def _is_space_with_block(cls, line):
        if cls._can_be_space(line[0]):
            if all(pixel in (BOX, UNKNOWN) for pixel in line[1:]):
                return True

        return False

    @classmethod
    def _space_with_block(cls, block_size):
        return [cls.empty_cell()] + ([BOX] * block_size)

    @classmethod
    def empty_cell(cls):
        """
        Represent a single line symbol that is empty (no color)
        """
        return SPACE

    def paint(self, i, j):
        """
        Paint unsolved cells of line[:i+1]
        using blocks from 0 to (j+1)-th as description
        """
        if i < 0:
            return []

        painted = self._paint_table[self._linear_index(i, j)]
        if painted is None:
            if j < 0:
                if all(map(self._can_be_space, self.line[:i + 1])):
                    painted = [self.empty_cell()] * (i + 1)
                else:
                    raise NonogramError('Excess cells found at the beginning')
            else:
                painted = self._paint(i, j)

            self._paint_table[self._linear_index(i, j)] = painted

        return painted

    def _paint(self, i, j):
        fix0 = self._fix0(i, j)
        fix1 = self._fix1(i, j)

        if fix0:
            if fix1:
                return self._paint_both(i, j)

            return self._paint0(i, j)

        if fix1:
            return self._paint1(i, j)

        raise NonogramError('Block %r not fixable at position %r' % (j, i))

    def _paint0(self, i, j):
        return self.paint(i - 1, j) + [self.empty_cell()]

    def _paint1(self, i, j):
        block_size = self.description[j]
        return self.paint(i - block_size - 1, j - 1) + self._space_with_block(block_size)

    @classmethod
    def _merge_iter(cls, line1, line2):
        for pixel1, pixel2 in zip(line1, line2):
            if pixel1 == pixel2:
                yield pixel1
            else:
                yield UNKNOWN

    def _paint_both(self, i, j):
        return list(self._merge_iter(
            self._paint0(i, j),
            self._paint1(i, j)
        ))

    def _solve(self):
        res = self.paint(len(self.line) - 1, len(self.description) - 1)
        if self.additional_space:
            res = res[1:]

        return res




# -*- coding: utf-8 -*-
"""Define nonogram solver that uses contradictions"""

import time
from collections import OrderedDict, defaultdict, deque
from itertools import product

from notetool.tool.log import logger
from six import iteritems
from six.moves import range

from notegame.games.nonogram.core import propagation
from notegame.games.nonogram.core.board import CellPosition, CellState
from notegame.games.nonogram.core.common import NonogramError
from notegame.games.nonogram.solver.base import cache_info
from notegame.games.nonogram.utils.iter import expand_generator
from notegame.games.nonogram.utils.priority_dict import PriorityDict

USE_CONTRADICTION_RESULTS = True

# the fewer possible colors for a cell
# the earlier it appears in a list of candidates
FEW_COLORS_FIRST = False

# use the position of a cell (number of neighbours and solution rate of row/column)
# to adjust its rate when choosing the next probe for DFS
ADJUST_RATE = True


class _SearchNode(object):
    def __init__(self, value):
        self.value = value
        self.children = OrderedDict()

    def to_dict(self):
        """Represent the given node as the dictionary of other nodes"""

        if not self.children:
            return self.value

        return {
            'value': self.value,
            'children': OrderedDict(
                (str(k), v.to_dict())
                for k, v in iteritems(self.children)
            )
        }


class Solver(object):
    """
    Solve the nonogram using contradictions and depth-first search
    """

    def __init__(self, board, max_solutions=None, timeout=None, max_depth=None):
        """
        :type board: Board
        """
        self.board = board

        self.max_solutions = max_solutions
        self.timeout = timeout
        if max_depth is None:
            # why 440?
            # I simply searched for some value that will be somehow bigger
            # than 434 (the maximum useful search depth reached on the test set so far),
            # but still reachable for recursion calls.
            #
            # https://webpbn.com/30654 can be solved only with max_depth > 429
            # and the maximum depth reached on this puzzle is 434.
            #
            # NB: in current implementation depth=444 fails with
            # 'RuntimeError: maximum recursion depth exceeded'
            self.max_depth = 440
        else:
            self.max_depth = max_depth

        self.depth_reached = 0
        self.start_time = None
        self.explored_paths = set()
        self.search_map = None

    def _add_search_result(self, path, score):
        if isinstance(score, float):
            score = round(score, 4)

        if not path:
            if not self.search_map:
                self.search_map = _SearchNode(score)
            else:
                self.search_map.value = score
            return

        current = self.search_map

        for i, node in enumerate(path):
            # node = (i + 1, node)
            if i == len(path) - 1:
                val = score
            else:
                val = None

            if node not in current.children:
                current.children[node] = _SearchNode(val)

            current = current.children[node]

        current.value = score

    def propagate_change(self, cell_state):
        """
        Set the given color to given cell
        and try to solve the board with that new info.
        :type cell_state: CellState
        """
        board = self.board
        logger.debug('Assume that (%i, %i) is %s',  *tuple(cell_state))

        board.set_color(cell_state)

        return propagation.solve(board,
                                 row_indexes=(cell_state.row_index,),
                                 column_indexes=(cell_state.column_index,),
                                 contradiction_mode=True)

    def probe(self, cell_state, rollback=True, force=False):
        """
        Try to find if the given cell can be in an assumed state.
        If the contradiction is found, set the cell
        in an inverted state and propagate the changes if needed.
        If `rollback` then the solved board will restore to the previous state
        after the assumption was made.
        If `force`, try to solve it anyway, even if the cell is already solved
        (to rerun the line solver).

        Return the pair `(is_contradiction, new_info)` where

        is_contradiction: whether the assumption led to a contradiction

        new_info:
          a) when contradiction is found, it contains the state of the board
          before any assumptions was made. It will help further to determine
          which cells has changed on that probe.
          b) if no contradiction found, but `rollback` is False,
          then we do not restore the board and return the previous state also.
          c) otherwise it contains the number of solved cells for the partially
          solved board (if the assumption made is true)
        """
        board = self.board

        pos = cell_state.position
        assumption = cell_state.color
        # already solved
        if board.is_cell_solved(pos):
            if not force:
                return False, None

        if assumption not in board.cell_colors(pos):
            logger.warning(
                "The probe is useless: color '%s' already unset" % (assumption))
            return False, None

        save = board.make_snapshot()

        try:
            solved_cells = self.propagate_change(cell_state)
        except NonogramError:
            logger.debug('Contradiction', exc_info=True)
            # rollback solved cells
            board.restore(save)

        else:
            if board.is_solved_full:
                self._add_solution()

            if rollback:
                board.restore(save)
                return False, solved_cells

            return False, save

        if USE_CONTRADICTION_RESULTS:
            before_contradiction = board.make_snapshot()
        else:
            before_contradiction = None

        pos = cell_state.position
        logger.info('Found contradiction at (%i, %i)' , *pos)
        try:
            board.unset_color(cell_state)
        except ValueError as ex:
            raise NonogramError(str(ex))

        # try to solve with additional info
        # solve with only one cell as new info
        propagation.solve(
            board,
            row_indexes=(pos.row_index,),
            column_indexes=(pos.column_index,))

        return True, before_contradiction

    def _new_jobs_from_solution(self, cell_state, previous_board, is_contradiction):
        board = self.board

        # evaluate generator
        changed = list(board.changed(previous_board))
        assumption = cell_state.color
        log_contradiction = '(not) ' if is_contradiction else ''
        logger.info('Changed %d cells with %s%s assumption' %
                    (len(changed), log_contradiction, assumption))

        # add the neighbours of the changed cells into jobs
        for pos in changed:
            for neighbour in board.unsolved_neighbours(pos):
                yield neighbour, 1

        # add the neighbours of the selected cell into jobs
        for neighbour in board.unsolved_neighbours(cell_state.position):
            yield neighbour, 0

    def _set_guess(self, state):
        board = self.board
        is_contradiction, prev_board = self.probe(
            state, rollback=False, force=True)

        if is_contradiction:
            raise NonogramError('Real contradiction was found: %s' % (state,))

        if prev_board is None:
            logger.warning(
                "The probe for state '%s' does not return anything new" % state)
            return ()

        if board.is_solved_full:
            self._add_solution()
            return ()

        return self._new_jobs_from_solution(state, prev_board, is_contradiction)

    def _solve_jobs(self, jobs, refill=False):
        """
        Given a board and a list of jobs try to solve that board
        using the jobs as probes.
        If `refill` specified, solve in several rounds,
        until all the contradictions disappear.

        Return the number of contradictions found and
        list of the best candidates for the tree-based search
        """

        counter, counter_found = 0, 0
        rates = dict()

        board = self.board

        processed_in_round = set()
        expired_assumptions = set()

        while jobs:
            state, priority = jobs.pop_smallest()
            counter += 1
            logger.info('Probe #%d: %s (%f)' % (counter, state, priority))

            # if the job is only coordinates
            # then try all the possible colors
            pos = state[:2]
            if len(state) == 2:
                assumptions = board.cell_colors(state)
            else:
                assumptions = (state.color,)

            for assumption in assumptions:
                state = CellState.from_position(pos, assumption)
                is_contradiction, info = self.probe(state)

                if info is None:
                    continue

                if is_contradiction:
                    counter_found += 1
                    if board.is_solved_full:
                        self._add_solution()
                        return counter_found, None

                    for new_job, priority in self._new_jobs_from_solution(
                            state, info, is_contradiction):
                        jobs[new_job] = priority

                    # All the jobs that was processed before the given
                    # contradiction considered expired. They will refill
                    # the active jobs queue when it gets empty.
                    expired_assumptions = set(processed_in_round)
                else:
                    rates[state] = (info, priority)

            if not refill:
                # only do one round of probes
                continue

            # we have work to do!
            if jobs:
                processed_in_round.add(pos)
            elif expired_assumptions:
                logger.warning('No more jobs. Refill all the jobs processed before '
                               'the last found contradiction (%s)' % (len(expired_assumptions)))
                refill_processed = self._get_all_unsolved_jobs(
                    choose_from_cells=expired_assumptions)
                for new_job, priority in iteritems(refill_processed):
                    jobs[new_job] = priority

                # they are no longer expired, as they active now!
                processed_in_round -= expired_assumptions
                expired_assumptions = set()
                counter = 0

        return counter_found, self._probes_from_rates(rates)

    def _add_solution(self):
        # force to check the board
        propagation.solve(self.board, contradiction_mode=True)
        self.board.add_solution()

    @expand_generator
    def _probes_from_rates(self, rates):
        jobs_with_rates = defaultdict(dict)

        for cell_state, (rate, priority) in iteritems(rates):
            pos = cell_state.position
            color = cell_state.color
            if self.board.is_cell_solved(pos):
                continue

            # the more priority the less desired that job
            # priority is in the range [0, 8]
            if ADJUST_RATE:
                rate += (10 - priority)
            # if rate > jobs_with_rates.get(job, 0):
            jobs_with_rates[pos][color] = rate

        max_rate = {pos: max(v.values())
                    for pos, v in iteritems(jobs_with_rates)}
        # the biggest rate appears first
        best = sorted(iteritems(max_rate), key=lambda x: x[1], reverse=True)
        if FEW_COLORS_FIRST:
            best = sorted(best, key=lambda x: len(jobs_with_rates[x[0]]))
        logger.debug('\n'.join(map(str, best)))

        for pos, max_rate in best:
            colors = sorted(
                iteritems(jobs_with_rates[pos]), key=lambda x: x[1], reverse=True)
            for color, rate in colors:
                yield CellState.from_position(pos, color)

    def _get_all_unsolved_jobs(self, choose_from_cells=None):
        board = self.board

        if choose_from_cells is None:
            # add every cell
            choose_from_cells = product(
                range(board.height), range(board.width))

        probe_jobs = PriorityDict()

        for pos in choose_from_cells:
            pos = CellPosition(*pos)
            if board.is_cell_solved(pos):
                continue

            no_unsolved = len(list(board.unsolved_neighbours(pos)))

            # if no_unsolved >= 4 and skip_low_rated:
            #     continue

            row_rate = board.row_solution_rate(pos.row_index)
            column_rate = board.column_solution_rate(pos.column_index)
            cell_rate = row_rate + column_rate

            probe_jobs[pos] = 4 - cell_rate + no_unsolved

        return probe_jobs

    def _solve_without_search(self, to_the_max=False):
        """
        Do the one round of solving with contradictions.
        Returns the number of contradictions found.

        Based on https://www.cs.bgu.ac.il/~benr/nonograms/
        """

        probe_jobs = self._get_all_unsolved_jobs()
        return self._solve_jobs(probe_jobs, refill=to_the_max)

    @classmethod
    def shrink_board(cls, board, candidates=()):
        """
        Skip the solved rows in a board before backtracking.
        Also fix candidate states.
        """
        solved_columns, solved_rows = board.reduce()

        if solved_rows and solved_rows[0]:
            skip_first_rows = len(solved_rows[0])

            for index, candidate in enumerate(candidates):  # type: CellState
                if candidate.row_index < skip_first_rows:
                    raise ValueError(
                        'Bad candidate: %r. The row skipped' % candidate)

                row_index, col_index, color = candidate
                new_candidate = CellState(
                    row_index - skip_first_rows, col_index, color)
                candidates[index] = new_candidate
                logger.info('Fixed candidate: %r -> %r' %
                            (candidate, new_candidate))

        if solved_columns and solved_columns[0]:
            skip_first_columns = len(solved_columns[0])

            for index, candidate in enumerate(candidates):  # type: CellState
                if candidate.column_index < skip_first_columns:
                    raise ValueError(
                        'Bad candidate: %r. The column skipped' % candidate)

                row_index, col_index, color = candidate
                new_candidate = CellState(
                    row_index, col_index - skip_first_columns, color)
                candidates[index] = new_candidate
                logger.info('Fixed candidate: %r -> %r',
                            candidate, new_candidate)

        return candidates

    @classmethod
    def _fix_candidates_colors(cls, candidates, color_mapping):
        for index, candidate in enumerate(candidates):  # type: CellState
            row_index, col_index, color = candidate
            if color not in color_mapping:
                logger.warning('Bad candidate %r', candidate)
                continue

            new_candidate = CellState(
                row_index, col_index, color_mapping[color])
            candidates[index] = new_candidate
            logger.info('Fixed candidate: %r -> %r', candidate, new_candidate)

        return candidates

    def solve(self, methods=None):
        """
        Solve the nonogram to the most with contradictions
        """

        board = self.board

        propagation.solve(board, methods=methods)
        if board.is_solved_full:
            board.set_finished()
            logger.info('No need to solve with contradictions')
            return

        logger.warning('Trying to solve using contradictions method')
        start = time.time()

        found_contradictions, best_candidates = self._solve_without_search(
            to_the_max=True)
        current_solution_rate = board.solution_rate
        logger.warning('Contradictions (found %d): %f',
                       found_contradictions, current_solution_rate)

        if current_solution_rate < 1:
            # if stalled with sophisticated selection of cells
            # do the brute force search
            logger.warning('Starting DFS (intelligent brute-force)')
            best_candidates = self.shrink_board(
                board, candidates=best_candidates)
            if board.is_colored:
                single_colored, color_mapping = board.reduce_to_single_color()
                if single_colored is not None:
                    logger.warning('Replacing colored board with the '
                                   'equivalent black and white: %r', color_mapping)

                    # from now we will search the black and white board
                    self.board = single_colored
                    best_candidates = self._fix_candidates_colors(
                        best_candidates, color_mapping)

            self.search(best_candidates)
            board.restore_reduced()

            current_solution_rate = board.solution_rate
            logger.warning('Search completed (depth reached: %d, solutions found: %d)',
                           self.depth_reached, len(board.solutions))

        if current_solution_rate != 1:
            logger.warning('The nonogram is not solved full (with contradictions). '
                           'The rate is %.4f', current_solution_rate)

        board.set_finished()
        logger.info('Full solution: %.6f sec', time.time() - start)
        for method, info in cache_info().items():
            size, hit_rate = info
            if size > 0:
                logger.warning('%s cache: size=%d, hit rate=%.4f%%',
                               method, size, hit_rate * 100.0)

    def _limits_reached(self, depth):
        """
        Whether we reached the defined limits:
        1) number of solutions found
        2) the maximum allowed run time
        """
        if self.max_solutions:
            solutions_number = len(self.board.solutions)
            if solutions_number >= self.max_solutions:
                if depth == 0:
                    # only show log on the top level
                    logger.warning('%d solutions is enough', solutions_number)

                return True

        if self.timeout and self.start_time:
            run_time = time.time() - self.start_time
            if run_time > self.timeout:
                if depth == 0:
                    # only show log on the top level
                    logger.warning('Searched too long: %.4fs', run_time)
                return True

        return False

    def _try_state(self, state, path):
        """
        Trying to search for solutions in the given direction.
        At first it set the given state and get a list of the
        further jobs for finding the contradictions.
        Later that jobs will be used as candidates for a deeper search.

        :param state: the next cell and color to set
        :param path: the path that already have done
        """
        board = self.board

        depth = len(path)
        full_path = path + (state,)

        # add every cell to the jobs queue
        probe_jobs = self._get_all_unsolved_jobs()
        try:
            # update with more prioritized cells
            for new_job, priority in self._set_guess(state):
                probe_jobs[new_job] = priority

            if self._limits_reached(depth):
                return True

            __, best_candidates = self._solve_jobs(probe_jobs)
        except NonogramError as ex:
            logger.warning('Dead end found (%s): %s', full_path[-1], str(ex))
            self._add_search_result(full_path, False)
            return False

        rate = board.solution_rate
        logger.info('Reached rate %.4f on %s path', rate, full_path)
        self._add_search_result(full_path, rate)

        if rate == 1 or self._limits_reached(depth):
            return True

        cells_left = round((1 - rate) * board.width * board.height)
        logger.info('Unsolved cells left: %d', cells_left)

        if best_candidates:
            return self.search(best_candidates, path=full_path)

        return True

    def _set_explored(self, path):
        self.explored_paths.add(tuple(sorted(path)))

    def _is_explored(self, path):
        return tuple(sorted(path)) in self.explored_paths

    def search(self, states, path=()):
        """
        Recursively search for solutions

        Return False if the given path is a dead end (no solutions can be found)
        """

        if self._is_explored(path):
            return True

        if self.start_time is None:
            self.start_time = time.time()

        board = self.board
        depth = len(path)

        if self._limits_reached(depth):
            return True

        if self.max_depth and depth >= self.max_depth:
            logger.warning('Next step on the depth %d is deeper than the max (%d)',
                           depth, self.max_depth)
            return True

        # going to dive deeper, so increment it (full_path's length)
        if depth + 1 > self.depth_reached:
            self.depth_reached = depth + 1

        search_directions = deque(states)

        unconditional = False
        search_counter = 0
        save = board.make_snapshot()
        try:
            while search_directions:
                total_number_of_directions = len(search_directions)
                state = search_directions.popleft()
                search_counter += 1

                if self._limits_reached(depth):
                    return True

                if state in path:
                    continue

                assumption = state.color
                pos = state.position
                cell_colors = board.cell_colors(pos)

                if assumption not in cell_colors:
                    logger.warning("The assumption '%s' is already expired. "
                                   "Possible colors for %s are %s",
                                   assumption, pos, cell_colors)
                    continue

                if len(cell_colors) == 1:
                    logger.warning('Only one color for cell %r left: %s. Solve it unconditionally',
                                   pos, assumption)
                    assert assumption == tuple(cell_colors)[0]
                    if unconditional:
                        logger.warning(
                            'The board does not change since the last unconditional solving, skip.')
                        continue

                    try:
                        self._solve_without_search()
                        unconditional = True
                    except NonogramError:
                        # the whole `path` branch of a search tree is a dead end
                        logger.warning(
                            "The last possible color '%s' for the cell '%s' "
                            "lead to the contradiction. "
                            "The path %s is invalid", assumption, pos, path)
                        # self._add_search_result(path, False)
                        return False

                    # rate = board.solution_rate
                    # self._add_search_result(path, rate)
                    if board.is_solved_full:
                        self._add_solution()
                        logger.warning(
                            "The only color '%s' for the cell '%s' lead to full solution. "
                            "No need to traverse the path %s anymore", assumption, pos, path)
                        return True
                    continue

                full_path = path + (state,)
                if self._is_explored(full_path):
                    logger.info('The path %s already explored', full_path)
                    continue

                unconditional = False
                rate = board.solution_rate
                guess_save = board.make_snapshot()
                try:
                    logger.warning('Trying state (%d/%d): %s (depth=%d, rate=%.4f, previous=%s)',
                                   search_counter, total_number_of_directions,
                                   state, depth, rate, path)
                    self._add_search_result(path, rate)
                    success = self._try_state(state, path)
                    # is_solved = board.is_solved_full
                finally:
                    board.restore(guess_save)
                    self._set_explored(full_path)

                if not success:
                    # TODO: add backjumping here
                    try:
                        logger.warning(
                            "Unset the color %s for cell '%s'. Solve it unconditionally",
                            assumption, pos)
                        board.unset_color(state)
                        self._solve_without_search()
                        unconditional = True
                    except ValueError:
                        # the whole `path` branch of a search tree is a dead end
                        logger.warning(
                            "The last possible color '%s' for the cell '%s' "
                            "lead to the contradiction. "
                            "The whole branch (depth=%d) is invalid. ", assumption, pos, depth)
                        # self._add_search_result(path, False)
                        return False

                    # rate = board.solution_rate
                    # self._add_search_result(path, rate)
                    if board.is_solved_full:
                        self._add_solution()
                        logger.warning(
                            "The negation of color '%s' for the cell '%s' lead to full solution. "
                            "No need to traverse the path %s anymore", assumption, pos, path)
                        return True

                if not success or board.is_solved_full:
                    # immediately try the other colors as well
                    # if all of them goes to the dead end,
                    # then the parent path is a dead end
                    states_to_try = []
                    for color in cell_colors:
                        if color == assumption:
                            continue

                        states_to_try.append(
                            CellState.from_position(pos, color))

                    # if all(self._is_explored(path + (state,)) for state in states_to_try):
                    #     LOG.warning('All other colors (%s) of cell %s already explored',
                    #                 states_to_try, cell)
                    #     return True

                    for state in states_to_try:
                        if state not in search_directions:
                            search_directions.appendleft(state)

        finally:
            # do not restore the solved cells on a root path - they are really solved!
            if path:
                board.restore(save)
                self._set_explored(path)

        return True



# -*- coding: utf-8 -*-

from collections import defaultdict
from functools import wraps
from time import time

from notetool.tool.log import logger


class Cache(object):
    """
    具有大小限制和命中计数器的简单词典。
    """

    def __init__(self, max_size=10 ** 5, increase=False, do_not_increase_after=10 ** 6):
        """
        :param max_size: maximum number of items that the cache can store
        :param increase: whether to increase the size dynamically when reached the max.
        If you specify the True, the size will simply doubled. If you specify a number,
        the size will be multiplied by that amount.
        :param do_not_increase_after: prevent the cache from growing
        at certain number of items
        """
        self._storage = dict()
        self.init_size = max_size
        self.max_size = max_size
        self.hits = 0
        self.total_queries = 0

        if increase is True:
            increase = 2
        self.increase = increase
        self.do_not_increase_after = do_not_increase_after

    def __len__(self):
        return len(self._storage)

    def _clear(self):
        self._storage.clear()

    def save(self, name, value, **kwargs):
        """Write the value to cache."""

        if len(self) >= self.max_size:
            logger.warning(
                'Maximum size for cache reached (%s).', self.max_size)
            self._clear()
            self._increase_size()

        self._save(name, value, **kwargs)

    # noinspection PyUnusedLocal
    def _save(self, name, value, **kwargs):
        self._storage[name] = value

    def get(self, name):
        """Get the value from a cache"""

        self.total_queries += 1
        value = self._get(name)

        if value is None:
            return None

        self.hits += 1
        return value

    def _get(self, name):
        return self._storage.get(name)

    def _increase_size(self):
        if self.max_size >= self.do_not_increase_after:
            return

        if self.increase and self.increase > 1:
            new_max = self.max_size * self.increase
            self.max_size = min(new_max, self.do_not_increase_after)
        else:
            logger.info('Bad increase multiplier: %s', self.increase)

    def delete(self, name):
        """Just drop the value from a cache"""
        return bool(self._storage.pop(name, False))

    @property
    def hit_rate(self):
        """How much queries successfully reached the cache"""
        if not self.total_queries:
            return 0

        return float(self.hits) / self.total_queries


# noinspection SpellCheckingInspection
# https://english.stackexchange.com/a/312087
class ExpirableCache(Cache):
    """
    The cache with limited support for expiration.
    """

    def _save(self, name, value, **kwargs):
        """Optionally you can specify an expiration timeout"""
        _time = kwargs.get('time')
        self._storage[name] = (time(), _time, value)

    def _get(self, name):
        item = self._storage.get(name)

        if item is None:
            return None

        start, _time, value = item
        # expires
        if _time is not None and time() - start > _time:
            self.delete(name)

        return value


def memoized_two_args(func, cache=None):  # pragma: no cover
    """
    Memoize results of two-argument function.
    The first argument should be more 'volatile' and the second one more 'constant'.
    """
    if cache is None:
        cache = defaultdict(dict)

    @wraps(func)
    def wrapper(arg1, arg2):
        arg2_cache = cache[arg2]
        try:
            return arg2_cache[arg1]
        except KeyError:
            arg2_cache[arg1] = value = func(arg1, arg2)
            return value

    return wrapper


def init_once(func):
    """
    Implements common behaviour

    def some_param(self):
        if self._some_param is None:
            self._some_param = ... # init code

        return self._some_param
    """

    func_name = func.__name__
    result_member = '__result_' + func_name

    @wraps(func)
    def wrapper(arg):
        self = arg
        try:
            return getattr(self, result_member)
        except AttributeError:
            res = func(self)

            # only save the result if `func` is an instance or class method
            if hasattr(arg, func_name):
                setattr(self, result_member, res)
            return res

    return wrapper



import os

import pandas as pd
from tqdm import tqdm


def load_tag_info(path):
    tag_info = open(path).read()
    tag_info = tag_info.replace('  ', ' ')
    tag_info = tag_info.replace('  ', ' ')
    tag_info = tag_info.replace('  ', ' ')
    tag_info = tag_info.replace(':', '')
    tag_info = tag_info.replace(' \n', '\n')
    tag_info_list = tag_info.split('\n')

    tmp = [[i for i in line.split(' ')] for line in tag_info_list[2:] if ' ' in line]

    tag_df = pd.DataFrame(tmp)

    tag_df.columns = ['tag_id', 'x', 'y', 'z']
    tag_df['tag_id'] = tag_df['tag_id'].astype('int')
    for key in ['x', 'y', 'z']:
        tag_df[key] = tag_df[key].astype('float')*10.0
    return tag_df


def load_distince_data_origin(path):
    d1 = open(path).read()
    d2 = d1.split('\n')
    d3 = pd.DataFrame([line.split(':') for line in d2 if len(line.split(':')) == 9])
    d3.columns = ['c1', 'unixtime', 'c3', 'tag_id', 'anchor_id', 'distance', 'distance_check', 'c8', 'data_index']
    d3['distance'] = d3['distance'].astype('float')
    d3['distance_check'] = d3['distance_check'].astype('float')

    return d3


def load_distince_data(path):
    d3 = load_distince_data_origin(path)

    d41 = d3[['data_index', 'anchor_id', 'distance']].pivot(index='data_index', columns='anchor_id', values='distance')
    d41.reset_index(inplace=True)
    d41.columns = ['data_index', 'dis_0', 'dis_1', 'dis_2', 'dis_3']

    d42 = d3[['data_index', 'anchor_id', 'distance_check']].pivot(
        index='data_index', columns='anchor_id', values='distance_check')
    d42.reset_index(inplace=True)
    d42.columns = ['data_index', 'dis_c_0', 'dis_c_1', 'dis_c_2', 'dis_c_3']

    d5 = d3[['c1', 'data_index', 'c3', 'tag_id', 'c8', 'unixtime']].groupby(['data_index']).max().reset_index()

    d6 = pd.merge(d5, d41, on=['data_index'])
    d6 = pd.merge(d6, d42, on=['data_index'])

    d6.reset_index(drop=True, inplace=True)

    assert len(d41) == len(d41) == len(d5) == len(d6)

    return d6


def load_all_and_merge(path_dir, target_file=None, overwrite=False):
    target_file = target_file or f'data/{os.path.basename(path_dir)}.csv'

    if not overwrite and os.path.exists(target_file):
        return pd.read_csv(target_file)

    file_list = os.listdir(path_dir)
    file_list = sorted(file_list, key=lambda x: int(x.split('.')[0]))

    dfs = []
    for file_name in tqdm(file_list):
        tag_id = int(file_name.split('.')[0])
        path = os.path.join(path_dir, file_name)
        df1 = load_distince_data(path)
        df1['tag_id'] = tag_id
        dfs.append(df1)

    res = pd.concat(dfs)
    res.to_csv(target_file, index=None)
    return res



# -*- coding: utf-8 -*-
"""
Here lie the utilities methods that does not depend on any domain
e.g. manipulations with collections or streams.
"""

import sys
from functools import partial, wraps
from itertools import islice


def merge_dicts(*dict_args, **kwargs):
    """
    Thanks Aaron Hall http://stackoverflow.com/a/26853961/
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    _type = kwargs.get('type', dict)
    # noinspection PyCallingNonCallable
    result = _type()
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def pad(coll, size, padding, left=True):
    """
    Pad the collection `coll` with `padding` items
    until it reaches the length of `size`.
    By default do padding in the beginning (on the left side).
    """
    padding_size = size - len(coll)
    if padding_size <= 0:
        return coll

    padding = [padding] * padding_size
    new_list = list(coll)
    return padding + new_list if left else new_list + padding


def interleave(list_a, list_b):
    """
    >>> interleave([0, 2, 4, 6], [1, 3, 5])
    [0, 1, 2, 3, 4, 5, 6]

    https://stackoverflow.com/a/7947461
    """
    size_a, size_b = len(list_a), len(list_b)
    if size_a - size_b not in (0, 1):
        raise ValueError(
            "The lists' sizes are too different: ({}, {})".format(size_a, size_b))
    res = [None] * (size_a + size_b)
    res[::2] = list_a
    res[1::2] = list_b
    return res


def max_safe(*args, **kwargs):
    """
    Returns max element of an iterable.

    Adds a `default` keyword for any version of python that do not support it
    """
    if sys.version_info < (3, 4):  # `default` supported since 3.4
        if len(args) == 1:
            arg = args[0]
            if 'default' in kwargs:
                default = kwargs.pop('default')
                if not arg:
                    return default

                # https://stackoverflow.com/questions/36157995#comment59954203_36158079
                arg = list(arg)
                if not arg:
                    return default

                # if the `arg` was an iterator, it's exhausted already
                # so use a new list instead
                return max(arg, **kwargs)

    return max(*args, **kwargs)


def list_replace(a_list, old, new):
    """
    Replaces every `old` item in `a_list` with `new` item
    """
    for i, item in enumerate(a_list):
        if item == old:
            a_list[i] = new


def split_seq(iterable, size):
    """
    Split `iterable` into chunks with specified `size`
    # http://stackoverflow.com/a/312467/

    :param iterable: any iterable
    :param size: chunk size
    :return: chunks one by one
    """
    iterator = iter(iterable)
    item = list(islice(iterator, size))
    while item:
        yield item
        item = list(islice(iterator, size))


def avg(iterable):
    """
    Return the average value of an iterable of numbers
    """

    # the iterable can be an iterator that gets exhausted
    # while `sum` and `len` will return 0
    list_copy = list(iterable)
    if not list_copy:
        return None

    return sum(list_copy) / len(list_copy)


def expand_generator(func=None, type_=list):
    """
    Escape the common list construction idiom by creating generator functions.
    You can also produce another iterable (e.g. tuple)

    Compare:
    def foo_list():                                  @expand_generator
        foo = []                                     def foo():
        while something:                                 while something:
            if something_more:                               if something_more:
                foo.append('X')          vs                      yield 'X'
            else:                                            else:
                foo.append('Y')                                  yield 'Y'

        return foo

    TODO: consider replacing with Andrey Petrov's `listify`
    https://github.com/shazow/unstdlib.py/blob/master/unstdlib/standard/list_.py#L149
    """
    if func is None:
        return partial(expand_generator, type_=type_)

    @wraps(func)
    def wrapper(*args, **kwargs):
        # noinspection PyArgumentList
        return type_(func(*args, **kwargs))

    return wrapper


def max_continuous_interval(int_list):
    """
    The size of the longest continuous
    number sequence in a list
    """
    int_list = sorted(int_list)

    if not int_list:
        return 0

    max_size = 1
    first, last = int_list[0], int_list[0]
    for number in int_list[1:]:
        if number == last + 1:
            last = number
        else:
            max_size = max(max_size, last - first + 1)  # save previous range
            first, last = number, number

    max_size = max(max_size, last - first + 1)
    return max_size



# -*- coding: utf-8 -*-
"""Define nonogram solving operations"""

from notetool.tool.log import logger
from six import add_metaclass, iteritems, itervalues

from ..core.common import SPACE_COLORED, BlottedBlock, NonogramError
from ..utils.cache import Cache
from ..utils.other import two_powers


class TwoLayerCache(Cache):
    """
    Special cache for storing nonogram line solutions.
    The keys are pairs (clue, partial_solution)
    """

    def __len__(self):
        return sum(len(lines) for lines in itervalues(self._storage))

    def _save(self, name, value, **kwargs):
        clue, prev_line = name
        if clue not in self._storage:
            self._storage[clue] = dict()

        self._storage[clue][prev_line] = value

    def _get(self, name):
        clue, prev_line = name
        clue_solutions = self._storage.get(clue)
        if clue_solutions is None:
            return None

        return clue_solutions.get(prev_line)

    def delete(self, name):
        clue, prev_line = name
        clue_solutions = self._storage.get(clue)
        if clue_solutions is None:
            return False

        return bool(clue_solutions.pop(prev_line, False))


class LineSolutionsMeta(type):
    """
    A metaclass for line solvers.
    It adds the solutions_cache to the solver class.
    """

    registered_caches = {}

    def __new__(mcs, *args, **kwargs):
        new_cls = super(LineSolutionsMeta, mcs).__new__(mcs, *args, **kwargs)
        new_cls.solutions_cache = Cache(increase=True)
        mcs.registered_caches[new_cls.__name__] = new_cls.solutions_cache
        return new_cls


def cache_info():
    """Cache size and hit rate for different solvers"""
    return {
        class_name: (len(cache), cache.hit_rate)
        for class_name, cache in iteritems(LineSolutionsMeta.registered_caches)
    }


@add_metaclass(LineSolutionsMeta)
class BaseLineSolver(object):
    """
    Basic line nonogram solver which provides
    facilities to save and extract solutions using cache
    """

    def __init__(self, description, line):
        self.description = description
        self.line = line

    @classmethod
    def _error_message(cls, description, line, additional_info=''):
        """Solve the line (or use cached value)"""
        description, line = tuple(description), tuple(line)
        return '{}: Failed to solve line {!r} with clues {!r}{}'.format(
            cls.__name__, line, description, additional_info)

    @classmethod
    def solve(cls, description, line):
        """Solve the line (or use cached value)"""
        if not line:
            return line

        description, line = tuple(description), tuple(line)

        solved = cls.solutions_cache.get((description, line))

        if solved is False:
            raise NonogramError(
                cls._error_message(description, line, additional_info=' (cached)'))

        if solved is not None:
            assert len(solved) == len(line)
            return solved

        solver = cls(description, line)
        try:
            solved = tuple(solver._solve())
        except NonogramError as ex:
            cls.save_in_cache((description, line), False)
            raise NonogramError(
                cls._error_message(description, line, additional_info=': {}'.format(ex)))

        assert len(solved) == len(line)
        cls.save_in_cache((description, line), solved)
        return solved

    @classmethod
    def save_in_cache(cls, original, solved):
        """
        Put the solution in local cache.
        Use solved=False to show that the line is not solvable.
        """
        cls.solutions_cache.save(original, solved)

    def _solve(self):
        """
        Override this function with actual solving algorithm.
        You should return something iterable (tuple, list or even generator!).
        """
        return self.line


class TrimmedSolver(BaseLineSolver):
    """Define some helpers for colored puzzle solvers"""

    @classmethod
    def prefix_size(cls, line, item):
        """
        How many items appear at the beginning of the line
        """
        for spaces_size, cell in enumerate(line):
            if cell != item:
                return spaces_size

        return len(line)

    @classmethod
    def _trim_spaces(cls, line):  # pragma: no cover
        space = SPACE_COLORED
        head_size = cls.prefix_size(line, space)
        tail_size = cls.prefix_size(list(reversed(line)), space)

        if tail_size:
            line = line[head_size: -tail_size]
        else:
            line = line[head_size:]

        space_tuple = (space,)

        return space_tuple * head_size, line, space_tuple * tail_size

    @classmethod
    def starting_solved(cls, description, line):
        """
        Trim off the solved cells from the beginning of the line.
        Also fix the description respectively.

        Return the pair (number of solved cells, number of solved blocks)
        """
        space = SPACE_COLORED

        block_index = 0
        last_block = len(description) - 1

        pos = 0
        last_pos = len(line) - 1

        while pos <= last_pos:
            # skip definite spaces
            if line[pos] == space:
                pos += 1
                continue

            cell = line[pos]
            cell_colors = two_powers(cell)
            if len(cell_colors) > 1:
                break

            color = cell_colors[0]

            if block_index > last_block:
                raise NonogramError(
                    'Bad block index {} '
                    'for description {!r}'.format(block_index, description))

            block = description[block_index]
            if color != block.color:
                raise NonogramError(
                    'Color {!r} at the position {!r} of the line {!r}'
                    'does not match the color of the corresponding block {!r} '
                    'in description {!r}'.format(
                        color, pos, line, block, description))

            size = block.size
            if size == BlottedBlock:

                end_pos = pos + 1
                while end_pos <= last_pos and line[end_pos] == color:
                    end_pos += 1

                if end_pos <= last_pos:
                    cell_colors = two_powers(line[end_pos])
                    # can't say definitely whether the blotted block ends here
                    if color in cell_colors:
                        # the partially solved blotted block can be reduced to one cell
                        pos = end_pos - 1
                        break

                pos = end_pos
                block_index += 1

            else:
                if pos + size > last_pos + 1:
                    raise NonogramError(
                        'The {}-th block {!r} cannot be allocated in the line {!r}'.format(
                            block_index, block, line))

                if line[pos: pos + size] != [color] * size:
                    break

                if block_index < last_block:
                    next_block = description[block_index + 1]
                    if next_block.color == color:
                        try:
                            if line[pos + size] != space:
                                break
                        except IndexError:
                            raise NonogramError(
                                'The next ({}-th) block {!r} '
                                'cannot be allocated in the line {!r}'.format(
                                    block_index + 1, next_block, line))
                pos += size
                block_index += 1

        return pos, block_index

    @classmethod
    def _trim_solved_blocks(cls, description, line):
        logger.info('Trying to trim off solved cells: %r, %r',
                    description, line)

        beg_solved, beg_blocks = cls.starting_solved(
            list(description), list(line))
        fin_solved, fin_blocks = cls.starting_solved(
            list(reversed(description)), list(reversed(line)))

        start = tuple(line[:beg_solved])

        if fin_solved:
            end = tuple(line[-fin_solved:])
            line = line[beg_solved: -fin_solved]
        else:
            end = ()
            # hack to do foo[a: b] == foo[a:]: just assign b = -len(foo)
            # fin_solved = -len(line)
            line = line[beg_solved:]

        if fin_blocks:
            description = description[beg_blocks: -fin_blocks]
        else:
            description = description[beg_blocks:]

        if not line and description:
            raise NonogramError(
                'Some blocks left after trimming the whole line')

        return description, line, (start, end)

    @classmethod
    def solve(cls, description, line):
        """Optimize line solving by trimming off the spaces and solved blocks"""
        # head, line, tail = cls._trim_spaces(line)
        # if not line:  # if consists only of spaces
        #     return head

        description, line, edges = cls._trim_solved_blocks(description, line)
        if any(edges):
            logger.info('Trimmed edges: %r, %r', *edges)
            logger.info('What is left after trimming: description %r and line %r',
                        description, line)
        if not line:  # all the cells are solved
            solved = edges[0]
        else:
            solved = super(TrimmedSolver, cls).solve(description, line)
            solved = edges[0] + solved + edges[1]

        # # restore the spaces
        # if head or tail:
        #     solved = head + solved + tail

        return solved



# -*- coding: utf-8 -*-
"""
The program's entry point
"""

import json

from notegame.games.nonogram.core.backtracking import Solver
from notegame.games.nonogram.core.board import make_board
from notegame.games.nonogram.core.renderer import BaseAsciiRenderer
from notegame.games.nonogram.reader import read_example


def solve(d_board, draw_final=False, draw_probes=False, **solver_args):
    """
    Wrapper for solver that handles errors and prints out the results
    """

    if not draw_final:
        d_board.on_solution_round_complete = lambda board: board.draw()
        if d_board.has_blots:
            d_board.on_row_update = lambda index, board: board.draw()
            d_board.on_column_update = lambda index, board: board.draw()

    solver = Solver(d_board, **solver_args)

    exc = False
    try:
        solver.solve()
    except BaseException:
        exc = True
        raise
    finally:
        if exc or draw_final:
            # draw the last solved cells
            d_board.draw()

        if not d_board.is_solved_full:
            d_board.draw_solutions()

        if draw_probes and solver.search_map:
            print(json.dumps(solver.search_map.to_dict(), indent=1))


def draw_solution(board_def, draw_final=False, **solver_args):
    """Solve the given board in terminal with animation"""
    d_board = make_board(*board_def, renderer=BaseAsciiRenderer)

    solve(d_board, draw_final=draw_final, **solver_args)


def main():
    columns = [[7], [1], [1], [1], [7], [0],
               [3], [1, 1, 1], [1, 1, 1], [2], [0],
               [6], [0],
               [6], [0],
               [3], [1, 1], [5], [1, 1], [3], [0],
               [5, 1], ]
    rows = [[1, 1, 1],
            [1, 1, 1, 1, 1],
            [1, 1, 2, 1, 1, 3, 1],
            [5, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 4, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 2, 1, 1, 3, 1], ]

    board_def = read_example("hello.txt")
    board_def = (columns, rows)
    draw_solution(board_def)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Nonogram solver using finite state machine

See source article (in russian):
http://window.edu.ru/resource/781/57781
"""

from notetool.tool.log import logger
from six import iteritems, itervalues
from six.moves import range

from ..core.common import (BOX, SPACE, SPACE_COLORED, UNKNOWN, NonogramError,
                           is_list_like, normalize_description, normalize_row)
from ..utils import fsm
from ..utils.cache import Cache
from ..utils.iter import expand_generator
from .base import BaseLineSolver
from .simpson import FastSolver


class NonogramFSM(fsm.FiniteStateMachine):
    """
    Represents a special class of a FSM
    used to solve a nonogram
    """

    def __init__(self, description, state_map):
        self.description = description
        initial_state = state_map[0][1]
        final = state_map[-1][1]
        super(NonogramFSM, self).__init__(
            initial_state, state_map, final=final)

    @classmethod
    def _space(cls):
        return SPACE

    @classmethod
    def _can_be_space(cls, cell):
        """
        Whether the cell can be a space
        """
        return cell in (UNKNOWN, cls._space())

    @classmethod
    def _can_be_empty(cls, row):
        """
        Whether the row can contain only spaces
        """
        return all(cls._can_be_space(cell) for cell in row)

    @classmethod
    def _optional_space(cls, state):
        return (state, cls._space()), state

    @classmethod
    def _required_color(cls, state, color):
        return (state, color), state + 1

    @classmethod
    def _required_box(cls, state):
        return cls._required_color(state, BOX)

    @classmethod
    def _required_space(cls, state):
        return cls._required_color(state, cls._space())

    INITIAL_STATE = 1

    @classmethod
    @expand_generator
    def state_map_from_description(cls, description):
        """
        Construct the machine's state map
        from the description
        given in a nonogram definition
        """
        logger.debug('Clues: %s', description)

        state_counter = cls.INITIAL_STATE

        prev_color = None
        for block in description:
            if is_list_like(block):
                value, color = block
            else:
                value, color = block, BOX

            # it SHOULD be a space before block
            # if it is not a first block AND the previous block was of the same color
            if prev_color == color:
                trans, state_counter = cls._required_space(state_counter)
                logger.debug('Add transition: %s -> %s', trans, state_counter)
                yield trans, state_counter

            # it CAN be multiple spaces before every block
            trans, state_counter = cls._optional_space(state_counter)
            logger.debug('Add transition: %s -> %s', trans, state_counter)
            yield trans, state_counter

            # the block of some color
            for _ in range(value):
                trans, state_counter = cls._required_color(
                    state_counter, color)
                logger.debug('Add transition: %s -> %s', trans, state_counter)
                yield trans, state_counter

            prev_color = color

        # at the end of the line can be optional spaces
        trans, state_counter = cls._optional_space(state_counter)
        logger.debug('Add transition: %s -> %s', trans, state_counter)
        yield trans, state_counter

    def partial_match(self, row):
        """
        Verify if the row of (possibly partly unsolved) cells
        can be matched against the current machine
        i.e. that the row can be a partial solution of a nonogram
        """
        row = normalize_row(row)

        # save the state in case of something change it
        save_state = self.current_state

        try:
            possible_states = {self.initial_state}

            for i, cell in enumerate(row):
                step_possible_states = []
                for state in possible_states:
                    if cell in (BOX, UNKNOWN):
                        logger.debug('Check the ability to insert BOX')

                        next_step = self.reaction(BOX, current_state=state)
                        if next_step is None:
                            logger.debug(
                                'Cannot go from state %r with BOX', state)
                        else:
                            step_possible_states.append(next_step)

                    if cell in (SPACE, UNKNOWN):
                        logger.debug('Check the ability to insert SPACE')

                        next_step = self.reaction(SPACE, current_state=state)
                        if next_step is None:
                            logger.debug(
                                'Cannot go from state %r with SPACE', state)
                        else:
                            step_possible_states.append(next_step)

                if not step_possible_states:
                    return False

                logger.debug('Possible states after step %s: %s',
                             i, step_possible_states)
                possible_states = set(step_possible_states)

            logger.debug('Possible states after full scan: %s',
                         possible_states)
            return self.final_state in possible_states
        finally:
            self._state = save_state

    def solve_with_partial_match(self, row):
        """
        Solve the nonogram `row` using the FSM and `self.partial_match` logic
        """
        original_row = normalize_row(row)

        if not self.description and self._can_be_empty(row):
            return (self._space(),) * len(row)

        # do not change original
        solved = list(original_row)
        for i, cell in enumerate(original_row):
            if cell in (BOX, SPACE):
                continue

            logger.debug('Trying to guess the %s cell', i)

            temp_row = list(original_row)
            temp_row[i] = BOX
            can_be_box = self.partial_match(temp_row)
            logger.debug('The %s cell can%s be a BOX',
                         i, '' if can_be_box else 'not')

            temp_row[i] = SPACE
            can_be_space = self.partial_match(temp_row)
            logger.debug('The %s cell can%s be a SPACE',
                         i, '' if can_be_space else 'not')

            if can_be_box:
                if not can_be_space:
                    solved[i] = BOX
            elif can_be_space:
                solved[i] = SPACE
            else:
                raise NonogramError(
                    'The {} cell ({}) in a row {!r} cannot be neither space nor box'.format(
                        i, cell, original_row))
        return solved

    @classmethod
    def _types_for_cell(cls, cell):
        if cell in (BOX, SPACE):
            return cell,

        if cell == UNKNOWN:
            return BOX, SPACE

        return ()

    def _make_transition_table(self, row):
        # for each read cell store a list of StepState
        # plus O-th for the state before any read cells
        transition_table = TransitionTable.with_capacity(len(row) + 1)
        transition_table.append_transition(0, self.initial_state)

        def _shift_one_cell(cell_type, trans_index,
                            previous_step_state, previous_state, desc_cell=None):
            if desc_cell:  # pragma: no cover
                logger.debug('Add states with %s transition', desc_cell)

            new_state = self.reaction(cell_type, previous_state)
            if new_state is None:
                if desc_cell:  # pragma: no cover
                    logger.debug('Cannot go from %s with the %s cell',
                                 previous_state, desc_cell)
            else:
                transition_table.append_transition(
                    trans_index, new_state, previous_step_state, cell_type)

        # optimize lookups
        _types_for_cell = self._types_for_cell

        for i, cell in enumerate(row):
            transition_index = i + 1

            for prev_state, prev in iteritems(transition_table[i]):
                for _type in _types_for_cell(cell):
                    _shift_one_cell(_type, transition_index,
                                    prev, prev_state)

        return transition_table

    @classmethod
    def _cell_value_from_solved(cls, states):
        assert states

        if len(states) == 1:
            return states[0]

        return UNKNOWN

    def solve_with_reverse_tracking(self, row):
        """
        Solve the nonogram `row` using the FSM and reverse tracking
        """
        if not self.description and self._can_be_empty(row):
            return (self._space(),) * len(row)

        transition_table = self._make_transition_table(row)

        # print(transition_table)
        if self.final_state not in transition_table[-1]:
            raise NonogramError('Bad transition table: final state {!r} not found'.format(
                self.final_state))

        reversed_states = list(
            transition_table.reverse_tracking(self.final_state))
        solved_row = (
            self._cell_value_from_solved(states)
            for states in reversed(reversed_states))

        return solved_row

    def match(self, word):
        # special case for an empty description
        if not self.description:
            word = tuple(set(word))
            if len(word) == 1:
                space = word[0]
                if space == self._space():
                    return True

        return super(NonogramFSM, self).match(word)


class _StepState(object):
    """
    Stores state of the machine,
    the link to the previous StepState object
    and the list of transitions which led to this state
    """

    __slots__ = ['state', 'previous_states']

    def __init__(self, state, prev=None, cell_type=None):
        # self.id = id(self)
        self.state = state
        self.previous_states = dict()
        self.add_previous_state(prev, cell_type)

    def add_previous_state(self, prev, cell_type):
        """
        Add transition which led to the current StepState.
        As a result it could be one or more pairs (StepState, cell_type).
        """
        if prev is not None:
            self.previous_states[prev] = cell_type

    def __str__(self):
        previous_states = sorted(
            iteritems(self.previous_states),
            key=lambda x: x[0].state)

        return '({}): [{}]'.format(
            # self.id,
            self.state,
            ', '.join('{}<-{}'.format(prev.state, cell_type)
                      for prev, cell_type in previous_states))


class TransitionTable(list):
    """
    Stores the map of all the transitions that
    can be made with given automaton for given input row
    """

    @classmethod
    def with_capacity(cls, capacity):
        """Generates a list of dicts with given capacity"""
        return TransitionTable([dict() for _ in range(capacity)])

    def append_transition(self, cell_index, state, prev=None, cell_type=None):
        """
        Add transition that shifts from `prev` StepState
        after reading `cell_index`-th cell from a row
        (which value is `cell_type`)
        to the new StepState of given `state`
        """
        transition_row = self[cell_index]
        if state in transition_row:
            # LOG.debug('State %s already exists in transition table for cell %s',
            #           state, cell_index)
            transition_row[state].add_previous_state(prev, cell_type)
        else:
            transition_row[state] = _StepState(state, prev, cell_type)

    def __str__(self):
        res = []
        for i, states in enumerate(self):
            if i > 0:
                res.append('')
            res.append(i)
            res.extend(sorted(itervalues(states), key=lambda x: x.state))
            # for state, step in iteritems(states):
            #     res.append('({}): {}'.format(state, step))

        return '\n'.join(map(str, res))

    def reverse_tracking(self, final_state):
        """
        Find all the possible reverse ways.

        Yield the actions that the machine can do
        on each step from final to initial.
        """
        possible_states = {final_state}

        # ignore the first row, it's for pre-read state
        for row in reversed(self[1:]):
            step_possible_cell_types = set()
            step_possible_states = set()

            for state in possible_states:
                step = row[state]
                for prev, cell_type in iteritems(step.previous_states):
                    step_possible_cell_types.add(cell_type)
                    step_possible_states.add(prev.state)

            possible_states = step_possible_states
            yield tuple(step_possible_cell_types)


class BaseMachineSolver(BaseLineSolver):
    """
    Nonogram line solver that uses finite state machine
    """

    def __init__(self, description, line):
        super(BaseMachineSolver, self).__init__(description, line)
        self.nfsm = self.make_nfsm(description)

    NFSM_CLASS = NonogramFSM
    FSM_CACHE = Cache(1000)

    @classmethod
    def get_state_map(cls, description):
        """
        Construct a state map
        (collection of available FSM transitions)
        from the nonogram description.
        Use cached value if available.
        """
        nfsm_cls = cls.NFSM_CLASS

        # if description and is_list_like(description[0]):
        #     nfsm_cls = NonogramFSMColored

        state_map = cls.FSM_CACHE.get(description)
        if state_map is None:
            state_map = nfsm_cls.state_map_from_description(description)
            cls.FSM_CACHE.save(description, state_map)

        return state_map

    @classmethod
    def make_nfsm(cls, *description):
        """
        Produce the black-and-white or colored
        finite state machine for nonogram solving
        """

        if len(description) == 1:
            description = description[0]
        description = normalize_description(description)

        state_map = cls.get_state_map(description)

        return cls.NFSM_CLASS(description, state_map)


class PartialMatchSolver(BaseMachineSolver):
    """
    FSM Nonogram solver that uses 'partial match' method (slow)
    """

    def _solve(self):
        return self.nfsm.solve_with_partial_match(self.line)


class BaseReverseTrackingSolver(BaseMachineSolver):
    """
    FSM Nonogram solver that uses 'reverse tracking' method (fast)
    """

    def _solve(self):
        return self.nfsm.solve_with_reverse_tracking(self.line)


class ReverseTrackingSolver(BaseReverseTrackingSolver):
    """
    FSM Nonogram solver for black and white puzzles
    """

    @classmethod
    def save_in_cache(cls, original, solved):
        super(ReverseTrackingSolver, cls).save_in_cache(original, solved)

        # it's a complete solution, so other solvers can use it too
        FastSolver.save_in_cache(original, solved)


def assert_match(row_desc, row):
    """
    Verifies that the given row matches the description
    """

    nfsm = BaseMachineSolver.make_nfsm(row_desc)
    if not nfsm.match(row):
        raise NonogramError(
            'The row {!r} cannot fit in clue {!r}'.format(row, row_desc))



# -*- coding: utf-8 -*-
"""
Defines methods to parse data file with the board defined
"""

import os
import re

from six import PY2, string_types
from six.moves.configparser import RawConfigParser

from notegame.games.nonogram.core.color import ColorMap

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

_INLINE_COMMENT_PREFIXES = '#;'


def parse_line(description, inline_comments=_INLINE_COMMENT_PREFIXES):
    """
    Parse a line and correctly add the description(s) to a collection
    """

    # manually strip out the comments
    # py2 cannot ignore comments on a continuation line
    # https://stackoverflow.com/q/9110428/1177288
    #
    # PY3 can do it for you with 'inline_comment_prefixes' = '#;'
    if PY2:
        for comment_prefix in inline_comments:
            pos = description.find(comment_prefix)
            if pos != -1:
                # comment line or inline comment (after a space)
                if pos == 0 or description[pos - 1].isspace():
                    description = description[:pos]

        if not description:
            return None

    # there can be trailing commas if you copy from source code
    descriptions = description.strip(',').split(',')

    # strip all the spaces and quotes
    descriptions = [desc.strip().strip("'").strip('"').strip()
                    for desc in descriptions]
    return descriptions


def example_file(file_name=''):
    """
    Returns a path to the examples board in text files
    """
    examples_dir = os.path.join(CURRENT_DIR, 'examples')
    if not file_name:
        return examples_dir

    if os.path.isfile(file_name):
        return file_name

    file_name = os.path.join(examples_dir, file_name)
    if os.path.isfile(file_name):
        return file_name

    txt_file_name = file_name + '.txt'
    if os.path.isfile(txt_file_name):
        return txt_file_name

    # just return the original file name, don't know where is it
    return file_name


def read_example(board_file):
    """Return the board definition for given example name"""
    return read_ini(example_file(board_file))


class MultiLineConfigParser(RawConfigParser, object):
    """
    INI-file parser that allows multiple lines in a value
    to be treated like a list.
    Also adds the ';'-style inline comments (disabled in PY3)

    https://stackoverflow.com/a/11866695/1177288
    """

    def __init__(self, *args, **kwargs):
        # allow '#' or ';' as the start of a comment
        if not PY2 and 'inline_comment_prefixes' not in kwargs:
            kwargs['inline_comment_prefixes'] = _INLINE_COMMENT_PREFIXES

        # noinspection PyArgumentList
        super(MultiLineConfigParser, self).__init__(*args, **kwargs)

    def get_list(self, section, option):
        """Split the value into list, remove empty items"""
        value = self.get(section, option)
        return [x.strip() for x in value.splitlines() if x]


_COLOR_RE = re.compile(r'\((.+)\) (.+)')


def read_ini(content):
    """Return the board definition from an INI-file"""

    parser = MultiLineConfigParser()

    if isinstance(content, string_types):
        content = open(content)

    if PY2:
        # it's not deprecated for python2
        # noinspection PyDeprecation
        parser.readfp(content)  # pylint: disable=deprecated-method
    else:
        # readfp is deprecated in future versions
        parser.read_file(content)

    columns = []
    for col in parser.get_list('clues', 'columns'):
        col = parse_line(col)
        if col is not None:
            columns.extend(col)

    rows = []
    for row in parser.get_list('clues', 'rows'):
        row = parse_line(row)
        if row is not None:
            rows.extend(row)

    res = [columns, rows]

    if parser.has_section('colors'):
        colors = ColorMap()
        for color_name, color_desc in parser.items('colors'):
            match = _COLOR_RE.match(color_desc)
            # TODO: spit some info if not matched

            colors.make_color(color_name, *match.groups())

        if not colors.black_and_white:
            res.append(colors)

    return tuple(res)



# -*- coding: utf-8 -*-

import copy
from queue import LifoQueue, Queue

import numpy as np
from notetool.tool.log import logger


class Recorder(object):
    point = None  # 进行猜测的点
    point_index = 0  # 猜测候选列表使用的值的索引
    value = None  # 回溯记录的值


class Sudoku(object):
    def __init__(self, data):
        # 数据初始化(二维的object数组)
        self.value = np.array([[0] * 9] * 9, dtype=object)  # 数独的值，包括未解决和已解决的
        self.new_points = Queue()  # 先进先出，新解（已解决值）的坐标
        self.recorder = LifoQueue()  # 先进后出，回溯器
        self.guess_times = 0  # 猜测次数

        # 九宫格的基准列表
        self.base_points = [[0, 0], [0, 3], [0, 6], [
            3, 0], [3, 3], [3, 6], [6, 0], [6, 3], [6, 6]]

        # 整理数据
        _data = np.array(data).reshape(9, -1)
        for r in range(0, 9):
            for c in range(0, 9):
                if _data[r, c]:  # if not Zero
                    # numpy default is int32, convert to int
                    self.value[r, c] = int(_data[r, c])
                    # 新的确认的值添加到列表中，以便遍历
                    self.new_points.put((r, c))
                    # logger.debug(f'init: answer={self.value[r, c]} at {(r, c)}')
                else:  # if Zero, guess no. is 1-9
                    self.value[r, c] = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    # 剔除数字
    def _cut_num(self, point):
        r, c = point
        val = self.value[r, c]

        # 行
        for i, item in enumerate(self.value[r]):
            if isinstance(item, list):
                if item.count(val):
                    item.remove(val)

                    # 判断移除后，是否剩下一个元素
                    if len(item) == 1:
                        self.new_points.put((r, i))  # 添加坐标到“已解决”列表
                        logger.debug('only one in row: answer={} at {}'.format(
                            self.value[r, i], (r, i)))
                        self.value[r, i] = item[0]

        # 列
        for i, item in enumerate(self.value[:, c]):
            if isinstance(item, list):
                if item.count(val):
                    item.remove(val)

                    # 判断移除后，是否剩下一个元素
                    if len(item) == 1:
                        self.new_points.put((i, c))
                        logger.debug('only one in col: answer={} at {}'.format(
                            self.value[i, c], (i, c)))
                        self.value[i, c] = item[0]

        # 所在九宫格(3x3的数组)
        b_r, b_c = map(lambda x: x // 3 * 3, point)  # 九宫格基准点
        for m_r, row in enumerate(self.value[b_r:b_r + 3, b_c:b_c + 3]):
            for m_c, item in enumerate(row):
                if isinstance(item, list):
                    if item.count(val):
                        item.remove(val)

                        # 判断移除后，是否剩下一个元素
                        if len(item) == 1:
                            r = b_r + m_r
                            c = b_c + m_c
                            self.new_points.put((r, c))
                            # logger.debug(f'only one in block: answer={self.value[r, c]} at {(r, c)}')
                            self.value[r, c] = item[0]

    # 同一行、列或九宫格中, List里，可能性只有一个的情况
    def _check_one_possible(self):
        # 同一行只有一个数字的情况
        for r in range(0, 9):
            # 只取出是这一行是List的格子
            values = list(filter(lambda x: isinstance(x, list), self.value[r]))

            for c, item in enumerate(self.value[r]):
                if isinstance(item, list):
                    for value in item:
                        if sum(map(lambda x: x.count(value), values)) == 1:
                            self.value[r, c] = value
                            self.new_points.put((r, c))
                            logger.debug(
                                'list val is only one in row: answer={} at {}'.format(self.value[r, c], (r, c)))
                            return True

        # 同一列只有一个数字的情况
        for c in range(0, 9):
            values = list(
                filter(lambda x: isinstance(x, list), self.value[:, c]))

            for r, item in enumerate(self.value[:, c]):
                if isinstance(item, list):
                    for value in item:
                        if sum(map(lambda x: x.count(value), values)) == 1:
                            self.value[r, c] = value
                            self.new_points.put((r, c))
                            logger.debug(
                                'list val is only one in col: answer={} at {}'.format(self.value[r, c], (r, c)))
                            return True

        # 九宫格内的单元格只有一个数字的情况
        for r, c in self.base_points:
            # reshape: 3x3 改为1维数组
            values = list(filter(lambda x: isinstance(x, list),
                                 self.value[r:r + 3, c:c + 3].reshape(1, -1)[0]))

            for m_r, row in enumerate(self.value[r:r + 3, c:c + 3]):
                for m_c, item in enumerate(row):
                    if isinstance(item, list):
                        for value in item:
                            if sum(map(lambda x: x.count(value), values)) == 1:
                                self.value[r + m_r, c + m_c] = value
                                self.new_points.put((r + m_r, c + m_c))
                                logger.debug('list val is only one in block: answer={} at {}'.format(
                                    self.value[r + m_r, c + m_c], (r + m_r, c + m_c)))
                                return True

    # 同一个九宫格内数字在同一行或同一列处理(同行列隐性排除)
    def _check_same_num(self):
        for b_r, b_c in self.base_points:
            block = self.value[b_r:b_r + 3, b_c:b_c + 3]

            # 判断数字1~9在该九宫格的分布情况
            _data = block.reshape(1, -1)[0]
            for i in range(1, 10):
                result = map(lambda x: 0 if not isinstance(x[1], list) else x[0] + 1 if x[1].count(i) else 0,
                             enumerate(_data))
                result = list(filter(lambda x: x > 0, result))
                r_count = len(result)

                if r_count in [2, 3]:
                    # 2或3个元素才有可能同一行或同一列
                    rows = list(map(lambda x: (x - 1) // 3, result))
                    cols = list(map(lambda x: (x - 1) % 3, result))

                    if len(set(rows)) == 1:
                        # 同一行，去掉其他行的数字
                        result = list(map(lambda x: b_c + (x - 1) % 3, result))
                        row = b_r + rows[0]

                        for col in range(0, 9):
                            if col not in result:
                                item = self.value[row, col]
                                if isinstance(item, list):
                                    if item.count(i):
                                        item.remove(i)

                                        # 判断移除后，是否剩下一个元素
                                        if len(item) == 1:
                                            self.new_points.put((row, col))
                                            logger.debug(
                                                'block compare row: answer={} at {}'.format(self.value[row, col],
                                                                                            (row, col)))
                                            self.value[row, col] = item[0]
                                            return True

                    elif len(set(cols)) == 1:
                        # 同一列
                        result = list(
                            map(lambda x: b_r + (x - 1) // 3, result))
                        col = b_c + cols[0]

                        for row in range(0, 9):
                            if row not in result:
                                item = self.value[row, col]
                                if isinstance(item, list):
                                    if item.count(i):
                                        item.remove(i)

                                        # 判断移除后，是否剩下一个元素
                                        if len(item) == 1:
                                            self.new_points.put((row, col))
                                            logger.debug(
                                                'block compare col: answer={} at {}'.format(self.value[row, col],
                                                                                            (row, col)))
                                            self.value[row, col] = item[0]
                                            return True

    # 排除法解题
    def sudo_exclude(self):
        is_run_same = True
        is_run_one = True

        while is_run_same:
            while is_run_one:
                # 剔除数字
                while not self.new_points.empty():
                    point = self.new_points.get()  # 先进先出
                    self._cut_num(point)

                # 检查List里值为单个数字的情况，如有新answer则加入new_points Queue，立即_cut_num
                is_run_one = self._check_one_possible()

            # 检查同行或列的情况
            is_run_same = self._check_same_num()
            is_run_one = True

    # 得到有多少个确定的数字
    def get_num_count(self):
        return sum(map(lambda x: 1 if isinstance(x, int) else 0, self.value.reshape(1, -1)[0]))

    # 评分，找到最佳的猜测坐标
    def get_best_point(self):
        best_score = 0
        best_point = (0, 0)

        for r, row in enumerate(self.value):
            for c, item in enumerate(row):
                point_score = self._get_point_score((r, c))
                if best_score < point_score:
                    best_score = point_score
                    best_point = (r, c)
        return best_point

    # 计算某坐标的评分
    def _get_point_score(self, point):
        # 评分标准 (10-候选个数) + 同行确定数字个数 + 同列确定数字个数
        r, c = point
        item = self.value[r, c]

        if isinstance(item, list):
            score = 10 - len(item)
            score += sum(map(lambda x: 1 if isinstance(x, int)
                             else 0, self.value[r]))
            score += sum(map(lambda x: 1 if isinstance(x, int)
                             else 0, self.value[:, c]))
            return score
        else:
            return 0

    # 验证有没错误
    def check_value(self):
        # 行
        r = 0
        for row in self.value:
            nums = []
            lists = []
            for item in row:
                (lists if isinstance(item, list) else nums).append(item)
            if len(set(nums)) != len(nums):
                # logger.error(f'verify failed. dup in row {r}')
                logger.debug('verify failed. dup in row {}'.format(r))
                return False  # 数字要不重复
            if len(list(filter(lambda x: len(x) == 0, lists))):
                return False  # 候选列表不能为空集
            r += 1

        # 列
        for c in range(0, 9):
            nums = []
            lists = []
            col = self.value[:, c]

            for item in col:
                (lists if isinstance(item, list) else nums).append(item)
            if len(set(nums)) != len(nums):
                logger.debug('verify failed. dup in col {}'.format(c))
                return False  # 数字要不重复
            if len(list(filter(lambda x: len(x) == 0, lists))):
                return False  # 候选列表不能为空集

        # 九宫格
        for b_r, b_c in self.base_points:
            nums = []
            lists = []
            block = self.value[b_r:b_r + 3, b_c:b_c + 3].reshape(1, -1)[0]

            for item in block:
                (lists if isinstance(item, list) else nums).append(item)
            if len(set(nums)) != len(nums):
                logger.debug(
                    'verify failed. dup in block {}'.format((b_r, b_c)))
                return False  # 数字要不重复
            if len(list(filter(lambda x: len(x) == 0, lists))):
                return False  # 候选列表不能为空集
        return True

    # 猜测记录
    def record_guess(self, point, index=0):
        # 记录
        recorder = Recorder()
        recorder.point = point
        recorder.point_index = index
        # recorder.value = self.value.copy() #numpy的copy不行
        recorder.value = copy.deepcopy(self.value)
        self.recorder.put(recorder)
        logger.debug('added to LIFO queue: {}'.format(
            [x.point for x in self.recorder.queue]))
        self.guess_times += 1  # 记录猜测次数

        # 新一轮的排除处理
        item = self.value[point]
        # assume only 1 in this point
        self.value[point] = item[index]
        self.new_points.put(point)
        logger.debug(
            'guessing: answer={}/{} @{}'.format(self.value[point], item, point))
        self.sudo_exclude()

    # 回溯，需要先进后出
    def recall(self):
        while True:
            if self.recorder.empty():
                raise Exception('Sudo is wrong, no answer!')
            else:
                recorder = self.recorder.get()
                point = recorder.point
                index = recorder.point_index + 1
                item = recorder.value[point]

                # 判断索引是否超出范围
                # if not exceed，则再回溯一次
                if index < len(item):
                    break
                # if exceed, pop next recorder
                logger.debug(f'Recall! Try previous point.')

        logger.debug(
            'Recall! Try next possible in same point, {} @{}'.format(item[index], point))
        self.value = recorder.value
        self.record_guess(point, index)

    # Main function 解题
    def sudo_solve(self):
        # 第一次解题，排除法
        logger.debug('excluding knowning answers')
        self.sudo_exclude()
        logger.debug('excluded, current result:\n{}'.format(self.value))

        # 检查有没错误的，有错误的则回溯；没错误却未解开题目，则再猜测
        while True:
            if self.check_value():
                fixed_answer = self.get_num_count()
                logger.debug(
                    'current no. of fixed answers: {}'.format(fixed_answer))
                if fixed_answer == 81:
                    break
                else:
                    # 获取最佳猜测点
                    point = self.get_best_point()
                    logger.debug(
                        'Adding new guessing in LIFO, {}'.format(point))

                    # 记录并处理
                    self.record_guess(point)
                    logger.debug(
                        'guessed, current result:\n{}'.format(self.value))
            else:
                # 出错，则回溯，尝试下一个猜测
                self.recall()


def sudoku_generate(mask_rate=0.5):
    while True:
        n = 9
        m = np.zeros((n, n), np.int)
        rg = np.arange(1, n + 1)
        m[0, :] = np.random.choice(rg, n, replace=False)
        try:
            for r in range(1, n):
                for c in range(n):
                    col_rest = np.setdiff1d(rg, m[:r, c])
                    row_rest = np.setdiff1d(rg, m[r, :c])
                    avb1 = np.intersect1d(col_rest, row_rest)
                    sub_r, sub_c = r // 3, c // 3
                    avb2 = np.setdiff1d(np.arange(0, n + 1),
                                        m[sub_r * 3:(sub_r + 1) * 3, sub_c * 3:(sub_c + 1) * 3].ravel())
                    avb = np.intersect1d(avb1, avb2)
                    m[r, c] = np.random.choice(avb, size=1)
            break
        except ValueError:
            pass

    mm = m.copy()
    mm[np.random.choice([True, False], size=m.shape, p=[
                        mask_rate, 1 - mask_rate])] = 0

    return mm


def sudoku_check_solution(m):
    if isinstance(m, list):
        m = np.array(m)
    elif isinstance(m, str):
        m = np.loadtxt(m, dtype=np.int, delimiter=",")
    set_rg = set(np.arange(1, m.shape[0] + 1))
    no_good = False
    for i in range(m.shape[0]):
        for j in range(m.shape[1]):
            r1 = set(m[3 * (i // 3):3 * (i // 3 + 1), 3 * (j // 3)                       :3 * (j // 3 + 1)].ravel()) == set_rg
            r2 = set(m[i, :]) == set_rg
            r3 = set(m[:, j]) == set_rg
            if not (r1 and r2 and r3):
                no_good = True
                break
        if no_good:
            break
    if no_good:
        logger.info("Checked: not good")
    else:
        logger.info("Checked: OK")

    return not no_good


def sudoku_solve_solution1(array):
    sudo = Sudoku(array)
    sudo.sudo_solve()
    return sudo.value


def sudoku_solve_solution2(array):
    if isinstance(array, list):
        array = np.array(array)
    elif isinstance(array, str):
        array = np.loadtxt(array, dtype=np.int, delimiter=",")
    rg = np.arange(array.shape[0] + 1)
    while True:
        mt = array.copy()
        while True:
            d = []
            d_len = []
            for i in range(array.shape[0]):
                for j in range(array.shape[1]):
                    if mt[i, j] == 0:
                        possibles = np.setdiff1d(rg, np.union1d(np.union1d(mt[i, :], mt[:, j]),
                                                                mt[3 * (i // 3):3 * (i // 3 + 1),
                                                                   3 * (j // 3):3 * (j // 3 + 1)]))
                        d.append([i, j, possibles])
                        d_len.append(len(possibles))
            if len(d) == 0:
                break
            idx = np.argmin(d_len)
            i, j, p = d[idx]
            if len(p) > 0:
                num = np.random.choice(p)
            else:
                break
            mt[i, j] = num
            if len(d) == 0:
                break
        if np.all(mt != 0):
            break

    return mt


def sudoku_solve_solution(array, method=1):
    if method == 1:
        return sudoku_solve_solution1(array)
    else:
        return sudoku_solve_solution2(array)



import json
import time
from threading import Thread

from notegame.games.topwar.entity import ActionResponse
from notetool.secret import read_secret
from websocket import WebSocket, create_connection

ping_interval = 30


class ActionRequest:
    def __init__(self, cid=None, o=None, p=None):
        self.cid = cid
        self.o = o
        self.p = p or {}

    @staticmethod
    def hero_list():
        p = {}

        return ActionRequest(cid=861, o="4", p=p)

    @staticmethod
    def map_info(x=299, y=463, k=1554, width=15, height=30, march_info=True):
        p = {"x": x, "y": y, "k": k, "width": width, "height": height, "marchInfo": march_info}
        # {"c":1001,"o":"81","p":{"x":299,"y":463,"k":1554,"width":10,"height":19}}
        # {"c":901,"o":"77","p":{"x":299,"y":463,"k":1554,"width":15,"height":30,"marchInfo":true}}

        return ActionRequest(cid=901, o="77", p=p)

    def __str__(self):
        res = {"c": self.cid, "o": str(self.o), "p": json.dumps(self.p)}
        return json.dumps(res)


class TopWarAction:
    def __init__(self, token=None, web_socket=None):
        self.token = read_secret(value=token, cate1='notechats', cate2='notegame', cate3='topwar', cate4='user',
                                 cate5='token')
        self.web_socket = web_socket or "wss://server-knight-s1200.rivergame.net/s1554"
        self.ws: WebSocket = create_connection(self.web_socket)
        self.ws_init()

    def ws_init(self):
        print('connection established')
        print(f'token:{self.token}')
        data = {
            "c": 1,
            "o": "0",
            "p": {
                "token": self.token,
                "country": "CN",
                "lang": "zh_cn",
                "nationalFlag": 48,
                "ip": "0",
                "pf": "android",
                "platform": "webgame",
                "channel": "webgame_webgameCn",
                "platformVer": "1.218.3",
                "containerType": "web",
                "serverId": 1554,
                "serverInfoToken": self.token,
                "appVersion": "1.218.3",
                "gaid": "",
                "itemId": ""
            }
        }
        self.ws.send(json.dumps(data))

        def start_heartbeat():
            time.sleep(ping_interval)
            self.ws.send('2')
            start_heartbeat()

        Thread(target=start_heartbeat).start()

        def on_message():
            while True:
                msg = self.ws.recv()
                try:
                    response = ActionResponse(msg)
                    print(response)
                    if response.cid == 901:
                        print(response.data)
                except Exception as e:
                    print(f'error:{e}\t{msg}')

        Thread(target=on_message).start()


topwar = TopWarAction()

topwar.ws.send(ActionRequest.map_info().__str__())



# -*- coding: utf-8 -*-
"""
Defines the basic terms and functions for nonogram game
"""


from notetool.tool.log import logger
from six import integer_types, iteritems, string_types, with_metaclass
from six.moves import range

from notegame.games.nonogram.core.color import Color, ColorBlock
from notegame.games.nonogram.utils.iter import expand_generator, list_replace

# this cell has to be solved
UNKNOWN = None

# but boolean constants are even faster than Enums
BOX = True  # BlackAndWhite.BLACK
SPACE = False  # BlackAndWhite.WHITE

# for colored puzzles only integer is allowed
SPACE_COLORED = Color.white().id_

BLACK_AND_WHITE_COLORS = (UNKNOWN, SPACE, BOX)


class NonogramError(ValueError):
    """
    Represents an error occurred when trying
    to solve a nonogram which has an internal contradiction.
    """


def invert(cell_state):
    """
    Invert the given cell state:
    BOX --> SPACE
    SPACE --> BOX

    For other values return unchanged.
    """
    if cell_state == BOX:
        return SPACE

    if cell_state == SPACE:
        return BOX

    return cell_state


def normalize_description(row, color=False):
    """
    Normalize a nonogram description for a row to the standard tuple format:
    - empty value (None, 0, '', [], ()) becomes an empty tuple
    - tuple or list becomes simply the same tuple
    - single number becomes a tuple with one item
    - a string of space-separated numbers becomes a tuple of that numbers
    """
    if not row:  # None, 0, '', [], ()
        return ()

    if isinstance(row, (tuple, list)):
        return tuple(row)

    if isinstance(row, integer_types):
        return row,  # it's a tuple!

    if isinstance(row, string_types):
        blocks = row.split()
        if color:
            return tuple(blocks)
        return tuple(map(int, blocks))

    raise ValueError('Bad row: %s' % row)


INFORMAL_REPRESENTATIONS = {
    UNKNOWN: ('_', ' ', '?', '*'),
    SPACE: ('.', '0', 'O', '-'),
    BOX: ('X', '+', '#'),
}

FORMAL_ALPHABET = set(INFORMAL_REPRESENTATIONS)


def normalize_row(row):
    """
    Normalize an easy-to write row representation with a formal one
    """
    alphabet = set(row)
    if alphabet.issubset(FORMAL_ALPHABET):
        return row

    if is_color_list(row):
        return row

    logger.debug('All row symbols: %s', alphabet)
    # save original for logs and debug
    original, row = row, list(row)

    for formal, informal in iteritems(INFORMAL_REPRESENTATIONS):
        informal = set(informal) & alphabet
        if not informal:
            logger.debug('Not found %r in a row', formal)
            continue

        if len(informal) > 1:
            raise ValueError(
                "Cannot contain different representations '{}' "
                "of the same state '{}' in a single row '{}'".format(
                    ', '.join(sorted(informal)), formal, original))

        informal = informal.pop()
        logger.debug('Replace %r with a %r', informal, formal)
        list_replace(row, informal, formal)

    assert set(row).issubset(FORMAL_ALPHABET)
    return tuple(row)


def is_list_like(value):
    """Whether value is tuple or list"""
    return isinstance(value, (tuple, list))


def is_color_cell(value):
    """
    Whether the value is a combination of several colors
    """
    try:
        value + 1
    except TypeError:
        return False  # color is always integer

    return value not in BLACK_AND_WHITE_COLORS


def is_color_list(value):
    """
    Whether value is a list of black and white or colored puzzle

    If it's a colored row, the value will have values other than
    (UNKNOWN, SPACE=1, BOX=2)
    """

    # check `len` to handle both standard types and numpy arrays
    if len(value):
        return any(is_color_cell(item) for item in value)

    return False


def _line_clues(line, white_color_code):
    size = len(line)

    description = []
    colors = set()

    index = 0
    while index < size:
        block_begin = index
        color_number = line[index]

        while (index < size) and (line[index] == color_number):
            index += 1

        block_size = index - block_begin
        if (block_size > 0) and (color_number != white_color_code):
            colors.add(color_number)
            description.append(ColorBlock(block_size, color_number))

    return description, colors


def clues(solution_matrix, white_color_code=SPACE):
    """
    Generate nonogram description (columns and rows)
    from a solution matrix.
    """
    height = len(solution_matrix)
    if height == 0:
        return (), ()

    width = len(solution_matrix[0])
    if width == 0:
        return (), ()

    all_colors = set()

    columns = []
    for col_index in range(width):
        column = [solution_matrix[row_index][col_index]
                  for row_index in range(height)]
        description, colors = _line_clues(column, white_color_code)
        all_colors |= colors
        columns.append(description)

    rows = []
    for row_index in range(height):
        row = solution_matrix[row_index]
        description, colors = _line_clues(row, white_color_code)
        all_colors |= colors
        rows.append(description)

    # black and white, ignore colors
    if len(all_colors) == 1:
        columns = [[block.size for block in col] for col in columns]
        rows = [[block.size for block in r] for r in rows]

    return columns, rows


class _BlottedMeta(type):
    """Redefine __repr__ to reduce noise in logs"""

    def __repr__(cls):
        return 'BLOTTED'


class BlottedBlock(with_metaclass(_BlottedMeta, object)):
    """
    The size of the block is unknown
    (try to solve by hand https://webpbn.com/19407 to grasp the concept)
    Intentionally made not an integer to prevent treating it like a number.
    """

    @classmethod
    def how_many(cls, description):
        """The number of blotted blocks in the description row"""
        counter = 0

        if not is_list_like(description):
            return 0

        for block in description:
            if block == cls:
                counter += 1

            if is_list_like(block):
                if block[0] == cls:
                    counter += 1

        return counter

    @classmethod
    @expand_generator
    def replace_with_1(cls, description):
        """
        Every blotted block spans a minimum of 1 cell
        """
        for block in description:
            if is_list_like(block):
                if block.size == cls:
                    block = ColorBlock(1, block.color)

            elif block == cls:
                block = 1

            yield block

    @classmethod
    def matches(cls, description, line):
        """
        Whether the given solved line
        matches the blotted description
        """

        colored = is_color_list(line)

        white_color = SPACE_COLORED if colored else SPACE
        line_clues, __ = _line_clues(line, white_color_code=white_color)

        if len(description) != len(line_clues):
            return False

        for actual, from_line in zip(description, line_clues):
            if not colored:
                from_line = from_line[0]

            if actual == from_line:
                continue

            if actual == cls:
                if from_line > 0:
                    continue

            return False

        return True


@expand_generator
def partial_sums(blocks, colored=None):
    """
    Calculate the partial sum of the blocks
    """

    if not blocks:
        return

    if colored is None:
        colored = isinstance(blocks[0], ColorBlock)

    if colored:
        sum_so_far = blocks[0].size
        yield sum_so_far

        for i, block in enumerate(blocks[1:], 1):
            size, color = block
            sum_so_far += size

            if blocks[i - 1].color == color:
                # plus at least one space
                sum_so_far += 1

            yield sum_so_far
    else:
        sum_so_far = blocks[0]
        yield sum_so_far

        for block in blocks[1:]:
            sum_so_far += block + 1
            yield sum_so_far


def slack_space(line_size, desc):
    """How much space left when the line fully shifted to one edge"""
    if not desc:
        return line_size

    desc = BlottedBlock.replace_with_1(desc)
    sums = partial_sums(desc)
    min_line_size = sums[-1]
    return line_size - min_line_size



# -*- coding: utf-8 -*-
"""
Based on Matteo Dell'Amico's solution
https://gist.github.com/matteodellamico/4451520
"""

from heapq import heapify, heappush, heappop

from six import iteritems


class PriorityDict(dict):
    """
    Dictionary that can be used as a priority queue.

    Keys of the dictionary are items to be put into the queue, and values
    are their respective priorities. All dictionary methods work as expected.
    The advantage over a standard heapq-based priority queue is
    that priorities of items can be efficiently updated (amortized O(1))
    using code as 'the_dict[item] = new_priority.'

    The 'smallest' method can be used to return the object with lowest
    priority, and 'pop_smallest' also removes it.

    The 'sorted_iter' method provides a destructive sorted iterator.
    """

    def __init__(self, *args, **kwargs):
        super(PriorityDict, self).__init__(*args, **kwargs)
        self._rebuild_heap()

    def _rebuild_heap(self):
        self._heap = [(val, key) for key, val in iteritems(self)]
        heapify(self._heap)

    def smallest(self):
        """
        Return the item with the lowest priority.
        Raises IndexError if the object is empty.
        """

        heap = self._heap
        val, key = heap[0]
        while key not in self or self[key] != val:
            heappop(heap)
            val, key = heap[0]
        return key, val

    def pop_smallest(self):
        """
        Return the item with the lowest priority and remove it.

        Raises IndexError if the object is empty.
        """

        heap = self._heap
        val, key = heappop(heap)
        while key not in self or self[key] != val:
            val, key = heappop(heap)
        del self[key]
        return key, val

    def __setitem__(self, key, val):
        """
        We are not going to remove the previous value from the heap,
        since this would have a cost O(n).
        """

        super(PriorityDict, self).__setitem__(key, val)

        if len(self._heap) < 2 * len(self):
            heappush(self._heap, (val, key))
        else:
            # When the heap grows larger than 2 * len(self), we rebuild it
            # from scratch to avoid wasting too much memory.
            self._rebuild_heap()

    def setdefault(self, key, d=None):
        if key not in self:
            self[key] = d
            return d
        return self[key]

    def update(self, *args, **kwargs):
        """
        Reimplementing dict.update is tricky -- see e.g.
        http://mail.python.org/pipermail/python-ideas/2007-May/000744.html
        We just rebuild the heap from scratch after passing to super.
        """

        super(PriorityDict, self).update(*args, **kwargs)
        self._rebuild_heap()

    def sorted_iter(self):
        """
        Sorted iterator of the priority dictionary items.
        Beware: this will destroy elements as they are returned.
        """

        while self:
            yield self.pop_smallest()



# -*- coding: utf-8 -*-
"""
Dynamic programming algorithm to solve nonograms (using recursion)

See details:
https://www.cs.bgu.ac.il/~benr/nonograms/
"""

from itertools import product

from notetool.tool.log import logger
from six.moves import range, zip

from ..core.common import (BOX, SPACE, UNKNOWN, BlottedBlock, partial_sums,
                           slack_space)
from .base import BaseLineSolver, NonogramError

# dummy constant
BOTH_COLORS = -1


class BguSolver(BaseLineSolver):
    """
    The solver uses recursion to solve the line to the most
    """

    def __init__(self, description, line):
        super(BguSolver, self).__init__(description, line)
        self._additional_space = self._set_additional_space()

        self.block_sums = self.calc_block_sum(description)
        self.solved_line = list(self.line)

        self._reset_solutions_table()

    def _reset_solutions_table(self):
        positions = len(self.line)
        job_size = len(self.description) + 1
        self.sol = [[None] * positions for _ in range(job_size)]

    def _set_additional_space(self):
        """
        Define the internal representation of a line to be one cell larger then the original.
        This is done to avoid an edge case later in our recursive formula.
        """
        if self.line[-1] != SPACE:
            self.line = list(self.line) + [SPACE]
            return True

        return False

    def _solve(self):
        if self.try_solve():
            solved = self.solved_line
            if self._additional_space:
                solved = solved[:-1]
            solved = (UNKNOWN if cell ==
                      BOTH_COLORS else cell for cell in solved)
            return solved

        raise NonogramError('Bad line')

    def try_solve(self):
        """
        The main solver function.
        Return whether the line is solvable.
        """
        position, block = len(self.line) - 1, len(self.description)
        return self.get_sol(position, block)

    @classmethod
    def calc_block_sum(cls, blocks):
        """
        calculates the partial sum of the blocks.
        this is used later to determine if we can fit some blocks in the space left on the line
        """
        min_indexes = [s - 1 for s in partial_sums(blocks, colored=False)]
        return [0] + min_indexes

    def fill_matrix_top_down(self, position, block):
        """
        Calculate the solution for line[:position+1]
        in respect to description[:block]

        :param position: position of cell we're currently trying to fill
        :param block: current block number
        :return: whether the segment of line solvable
        """

        if (position < 0) or (block < 0):
            return None

        # too many blocks left to fit this line segment
        if position < self.block_sums[block]:
            return False

        # recursive case
        if self.line[position] == BOX:  # current cell is BOX
            return False  # can't place a block if the cell is black

        # base case
        if position == 0:  # reached the end of the line
            if block == 0:
                self.add_cell_color(position, SPACE)
                return True

            return False

        # current cell is either white or unknown
        white_ans = self.get_sol(position - 1, block)

        # block == 0 means we finished filling all the blocks (can still fill whitespace)
        if block > 0:
            block_size = self.description[block - 1]

            if self.can_place_block(position - block_size, block_size):
                black_ans = self.get_sol(position - block_size - 1, block - 1)
                if black_ans:
                    # set cell white, place the current block and continue
                    self.set_line_block(position - block_size, position)
                    return True

        if white_ans:
            # set cell white and continue
            self.add_cell_color(position, SPACE)
            return True

        return False  # no solution

    def can_place_block(self, position, length):
        """
        check if we can place a block of a specific length in this position
        we check that our partial solution does not negate the line's partial solution
        :param position:  position to place block at
        :param length:  length of block
        """
        if position < 0:
            return False

        # if no negations were found, the block can be placed
        return SPACE not in self.line[position: position + length]

    def add_cell_color(self, position, value):
        """sets a cell in the solution matrix"""

        cell = self.solved_line[position]
        if cell == BOTH_COLORS:
            pass
        elif cell == UNKNOWN:
            self.solved_line[position] = value
        elif cell != value:
            self.solved_line[position] = BOTH_COLORS

    def set_line_block(self, start_pos, end_pos):
        """
        sets a block in the solution matrix. all cells are painted black,
        except the end_pos which is white.
        :param start_pos: position to start painting
        :param end_pos: position to stop painting
        """

        # set blacks
        for i in range(start_pos, end_pos):
            self.add_cell_color(i, BOX)

        self.add_cell_color(end_pos, SPACE)

    def set_sol(self, position, block, value):
        """
        sets a value in the solution matrix
        so we wont calculate this value recursively anymore
        """
        if position < 0:
            return

        self.sol[block][position] = value

    def get_sol(self, position, block):
        """
        gets the value from the solution matrix
        if the value is missing, we calculate it recursively
        """

        if position == -1:
            # finished placing the last block, exactly at the beginning of the line.
            return block == 0

        # self._get_sol(position, block)
        can_be_solved = self.sol[block][position]
        if can_be_solved is None:
            can_be_solved = self.fill_matrix_top_down(position, block)

            # self.set_sol(position, block, can_be_solved)
            self.sol[block][position] = can_be_solved

        return can_be_solved


UNKNOWN_COLORED = 0


class BlottedSolver(BaseLineSolver):
    """Some additional routines to help solve blotted puzzles"""

    @classmethod
    def is_solved(cls, description, line):
        """
        Whether the line already solved.
        Do not solve if so, since the blotted algorithm is computationally heavy.
        """
        raise NotImplementedError()

    @classmethod
    def _blotted_combinations(cls, description, line):
        """
        Generate all the possible combinations of blotted blocks sizes.
        The routine suggests that every size can be in range [0..max_sum]
        """

        blocks_number = sum(1 for block in description
                            if cls._is_blotted(block))
        max_sum = slack_space(len(line), description)

        valid_range = range(max_sum + 1)
        for combination in product(valid_range, repeat=blocks_number):
            if sum(combination) <= max_sum:
                yield combination

    @classmethod
    def _is_blotted(cls, block):
        raise NotImplementedError()

    @classmethod
    def _update_block(cls, current, increase):
        raise NotImplementedError()

    @classmethod
    def _single_color(cls, values):
        raise NotImplementedError()

    @classmethod
    def merge_solutions(cls, one, other=None):
        """Merge solutions from different description suggestions"""
        if other is None:
            return one

        logger.debug('Merging two solutions: %r and %r', one, other)
        return [cls._single_color(set(cells))
                for cells in zip(one, other)]

    @classmethod
    def solve(cls, description, line):
        """Solve the line (or use cached value)"""
        if not line:
            return line

        if not BlottedBlock.how_many(description):
            return super(BlottedSolver, cls).solve(description, line)

        if cls.is_solved(description, line):
            logger.info('No need to solve blotted line: %r', line)
            return line

        blotted_desc, line = tuple(description), tuple(line)
        logger.warning('Solving line %r with blotted description %r',
                       line, blotted_desc)

        blotted_positions = [index for index, block in enumerate(blotted_desc)
                             if cls._is_blotted(block)]

        # prevent from incidental changing
        min_desc = tuple(BlottedBlock.replace_with_1(blotted_desc))

        solution = None
        for index, combination in enumerate(cls._blotted_combinations(
                blotted_desc, line)):

            current_description = list(min_desc)
            for pos, block_size in zip(blotted_positions, combination):
                block = current_description[pos]
                current_description[pos] = cls._update_block(block, block_size)

            logger.debug('Trying %i-th combination %r',
                         index, current_description)

            try:
                solved = tuple(super(BlottedSolver, cls).solve(
                    current_description, line))
            except NonogramError:
                logger.debug('Combination %r is invalid for line %r',
                             current_description, line)
            else:
                solution = cls.merge_solutions(solved, solution)
                logger.debug('Merged solution: %s', solution)
                if tuple(solution) == line:
                    logger.warning('The combination %r (description=%r) is valid but '
                                   'brings no new information. Stopping the combinations search.',
                                   combination, current_description)
                    break

        if not solution:
            raise NonogramError(
                'Cannot solve with blotted clues {!r}'.format(blotted_desc))

        logger.info('United solution from all combinations: %r', solution)
        assert len(solution) == len(line)
        return tuple(solution)


class BguBlottedSolver(BlottedSolver, BguSolver):
    """
    Slightly modified algorithm to solve with blotted descriptions
    """

    @classmethod
    def is_solved(cls, description, line):
        if UNKNOWN in line:
            return False

        return BlottedBlock.matches(description, line)

    @classmethod
    def calc_block_sum(cls, blocks):
        blocks = BlottedBlock.replace_with_1(blocks)
        return super(BguBlottedSolver, cls).calc_block_sum(blocks)

    @classmethod
    def _is_blotted(cls, block):
        return block == BlottedBlock

    @classmethod
    def _update_block(cls, current, increase):
        return current + increase

    @classmethod
    def _single_color(cls, values):
        if len(values) > 1:
            return UNKNOWN

        return tuple(values)[0]



# -*- coding: utf-8 -*-
"""
Here lie the utilities methods that does not depend on any domain
e.g. manipulations with collections or streams.
"""

import logging
import multiprocessing
import os
import sys
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from threading import Lock

from memoized import memoized
from six import iteritems, text_type

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
START_TIME = datetime.now()


def get_uptime():  # pragma: no cover
    """Return the time program run in human-readable form"""
    return text_type(datetime.now() - START_TIME)


_IMPORT_LOCK = Lock()


@contextmanager
def extend_import_path(dir_name, first=False):
    """
    Temporarily add the `dir_name` to the sys.path
    If `first` flag provided, the directory will be added
    at the very beginning to override any other imported resources.
    """

    if dir_name not in sys.path:
        with _IMPORT_LOCK:  # prevent from doing path manipulations concurrently
            if first:
                sys.path.insert(0, dir_name)
            else:
                sys.path.append(dir_name)

            logging.info('The %r added to the sys.path', dir_name)

            try:
                yield
            finally:
                if first:
                    removed_path = sys.path.pop(0)
                else:
                    removed_path = sys.path.pop()

                logging.info('The %r removed from the sys.path', removed_path)
                assert dir_name == removed_path


def get_version():
    """Return the program's version from the package root"""

    root_dir = os.path.dirname(CURRENT_DIR)

    with extend_import_path(root_dir, first=True):
        try:
            from __version__ import VERSION
        except ImportError:  # pragma: no cover
            # noinspection PyPep8Naming
            VERSION = ()

    return VERSION


@contextmanager
def terminating_mp_pool(*args, **kwargs):  # pragma: no cover
    """
    Allows to use multiprocessing.Pool as a contextmanager in both PY2 and PY3

    https://stackoverflow.com/a/25968716
    """
    pool = multiprocessing.Pool(*args, **kwargs)
    try:
        yield pool
    finally:
        pool.terminate()


def is_close(first, second, rel_tol=1e-09, abs_tol=0.0):
    """
    Almost equality for float numbers

    :param first
    :param second

    :param rel_tol: relative tolerance, it is multiplied
    by the greater of the magnitudes of the two arguments;
    as the values get larger, so does the allowed difference
    between them while still considering them equal.
    :param abs_tol: absolute tolerance that is applied as-is in all cases.

    If the difference is less than either of those tolerances,
    the values are considered equal.

    Source: https://stackoverflow.com/a/33024979
    """
    return abs(first - second) <= max(rel_tol * max(abs(first), abs(second)), abs_tol)


def log_call(log_func=print):  # pragma: no cover
    """Print every function call along with its arguments"""

    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__

            msg = 'Calling function {}('.format(func_name)
            _args = args
            if _args:
                _self = _args[0]
                if isinstance(_self, object) and hasattr(_self, func_name):
                    _args = args[1:]

            msg += ', '.join(map(str, _args))
            msg += ', '.join('{}={}'.format(k, v)
                             for k, v in iteritems(kwargs))

            msg += ')'

            # start_time = time.time()
            result = func(*args, **kwargs)
            msg += ' --> {}'.format(result)
            log_func(msg)

            return result

        return wrapper

    return _decorator


@memoized
def two_powers(num):
    """
    Get a 'factorization' of number into powers of 2:

    42 --> [2, 8, 32]
    """

    # https://stackoverflow.com/a/51786889/1177288
    # return tuple(1 << i for i, d in enumerate(reversed(bin(num)[2:])) if d == '1')
    return tuple(_two_powers(num))


def _two_powers(num):
    """
    https://stackoverflow.com/a/51859187/1177288
    """
    while num > 0:
        rest = num & (num - 1)
        yield num - rest
        num = rest


def from_two_powers(numbers):
    """
    Construct a number from the powers of 2
    """

    # TODO: consider a simple sum() here
    result = 0
    for num in numbers:
        result |= num

    return result


def get_named_logger(name__, file__, auto_config_when_main=True):
    """
    Choose the best name for logger based on how the file is called.
    When the file is run with interpreter like `python foo.py`,
    the logger will be named 'foo.py'.
    When the file imported by another module,
    the logger will be named 'parent_module.foo'.

    You should always call it with the __name__ and __file__ arguments.
    """
    if name__ == '__main__':  # pragma: no cover
        name__ = os.path.basename(file__)
        if auto_config_when_main:
            logging.basicConfig()

    return logging.getLogger(name__)


@contextmanager
def ignored(*exceptions):
    """https://www.youtube.com/watch?v=OSGv2VnC0go&t=2696s"""

    try:
        yield
    except exceptions:
        pass



import json
from datetime import datetime


class ActionRequest:
    def __init__(self, cid=None, o=None, p=None):
        self.cid = cid
        self.o = o
        self.p = p or {}

    @staticmethod
    def hero_list():
        p = {}

        return ActionRequest(cid=861, o="4", p=p)

    @staticmethod
    def map_info(x=100, y=100, k=1037, width=7, height=16, march_info=True):
        p = {"x": x, "y": y, "k": k, "width": width, "height": height, "marchInfo": march_info}
        # {"c":901,"o":"242","p":{"x":508,"y":384,"k":1037,"width":7,"height":16,"marchInfo":true}}

        return ActionRequest(cid=901, o="242", p=p)

    def __str__(self):
        res = {"c": self.cid, "o": str(self.o), "p": json.dumps(self.p)}
        return json.dumps(res)


class ActionResponse:
    def __init__(self, data=None):
        self.cid = None
        self.data = None
        self.time = None
        self.s = None
        self.o = None

        self.parse(data)

    def parse(self, data=None):
        if data is None:
            return

        data = json.loads(data)

        if len(set(data.keys()) - set('c,s,d,t,o'.split(','))) > 0:
            print(data.keys())

        self.cid = data['c']
        self.data = data['d']
        self.time = datetime.fromtimestamp(data['t'] / 1000.)
        self.s = data['s']
        self.o = data['o']

    def __str__(self):
        return f'{self.time}\t{self.cid}\t{self.s}\t{len(self.data)}'


class User:
    def __init__(self, msg_data=None):
        self.fan = None
        self.fat = None

        self.uid = None
        self.uuid = None
        self.nickname = None
        self.user_name = None
        self.user_gender = None

        self.parse_from_msg(msg_data)

    def parse_from_msg(self, data=None):
        if data is None:
            return
        self.fan = data['fan']
        self.fat = data['fat']
        self.uid = data['uid']
        self.uuid = data['uuid']
        self.nickname = data['fp']['nickname']
        self.user_name = data['fp']['username']
        self.user_gender = data['fp']['usergender']

    def __str__(self):
        return f'{self.uid}\t{self.uuid}\t{self.fat}\t{self.user_name}'


class MessageResponse:
    def __init__(self, data=None):
        self.content = None

        self.room_id = None
        self.time = None
        self.user = User()
        self.parse(data)

    def parse(self, data=None):
        data = json.loads(data)[1]
        data['fp'] = json.loads(data['fp'])
        self.time = datetime.fromtimestamp(data['time'] / 1000.)
        self.room_id = data['roomId']
        self.content = data['content']
        self.user.parse_from_msg(data)

    def __str__(self):
        return f'{self.time}\t{self.user}:\t{self.content}'



import itertools
import json

import numpy as np
from notegame.shumo.load_data import load_tag_info

masks = np.array([[0, 1, 1, 1],
                  [1, 0, 1, 1],
                  [1, 1, 0, 1],
                  [1, 1, 1, 0]])


class Anchor:
    def __init__(self, loc):
        self.loc = loc
        self.distance = None

    def init(self):
        self.distance = np.zeros([self.loc.shape[0], self.loc.shape[0]])
        for i in range(self.loc.shape[0]):
            for j in range(self.loc.shape[0]):
                self.distance[i][j] = round(np.linalg.norm(self.loc[i] - self.loc[j]), 5)

    @staticmethod
    def new_instance1():
        loc = np.array([[0, 0, 1300], [5000, 0, 1700],
                        [0, 5000, 1700], [5000, 5000, 1300]])
        anchor = Anchor(loc=loc)
        anchor.init()
        return anchor

    @staticmethod
    def new_instance2():
        loc = np.array([[0, 0, 1200], [5000, 0, 1600],
                        [0, 3000, 1600], [5000, 3000, 1200]])
        anchor = Anchor(loc=loc)
        anchor.init()
        return anchor


class TagInfoBak1:
    def __init__(self, path, anchor=None):
        self.tag_df = None
        self.anchor = anchor or Anchor.new_instance1()
        self.init(path)

    def init(self, path):
        self.tag_df = load_tag_info(path)

    def cul_distance(self):
        a = np.array(self.tag_df[['x', 'y', 'z']].values)
        for i in range(4):
            b = a - self.anchor.loc[i]
            self.tag_df[f'd{i}'] = np.round(np.linalg.norm(b, axis=1))


class TagInfo:
    def __init__(self, path, anchor=None):
        self.tag_df = None
        self.anchor = anchor or Anchor.new_instance1()
        self.init(path)

    def init(self, path):
        self.tag_df = load_tag_info(path)

    def cul_distance(self):
        a = np.array(self.tag_df[['x', 'y', 'z']].values)
        for i in range(4):
            b = a - self.anchor.loc[i]
            self.tag_df[f'd{i}'] = np.round(np.linalg.norm(b, axis=1))

    def check_data(self, tag_id, df0, normal=True):
        def check_field(df, field):
            value = df[field].values
            vmin = np.percentile(value, 20)
            vmax = np.percentile(value, 80)
            df2 = df[(df['dis_0'] >= vmin) & (df['dis_0'] <= vmax)]
            vmean = np.mean(df2['dis_0'])
            df.loc[(df['dis_0'] <= vmean - 20) | (df['dis_0'] >= vmean + 20), 'normal'] = 2

        def check(df):
            d1 = df[['normal', 'data_index', *[f'dis_{i}' for i in range(4)]]]

            # 异常值检测
            for index, line in enumerate(json.loads(d1.to_json(orient='records'))):
                is_normal = True
                if line['normal'] > 0:
                    continue
                # 看某条记录是佛有异常数据
                for i, j in itertools.combinations(np.arange(4), 2):
                    a, b, c = line[f'dis_{i}'], line[f'dis_{j}'], self.anchor.distance[i][j]

                    # 如果不满足三角形任意两边的和大于第三边
                    alpha = 0.95
                    beta = 0
                    if a * alpha - beta > b + c or b * alpha - beta > a + c or c * alpha - beta > a + b:
                        # print(i, j, a, b, c)
                        # is_normal = False
                        break

                if not is_normal:
                    df.loc[index, 'normal'] = 3

        df0 = df0.copy()

        if normal:
            for field in [f'dis_{i}' for i in range(4)]:
                check_field(df0, field)
        check(df0)
        return df0



import threading
import time

import websocket
from notegame.games.topwar.entity import MessageResponse
from notetool.secret import read_secret

ping_interval = 30


class TopWarMessage:
    def __init__(self, uuid=None, topwar_channels=None, web_socket=None):
        self.ws = None
        self.uuid = read_secret(cate1='notechats', cate2='notegame', cate3='topwar', cate4='user', cate5='uuid',
                                value=uuid)
        self.web_socket = web_socket or "wss://group-push.rivergame.net/socket.io/sjzb/1554/?EIO=3&transport=websocket"
        self.topwar_channels = topwar_channels or {"alliance_666": "102_2_1554_767182",
                                                   "world_1554": "102_1_1554",
                                                   "cross": "102_3_cross_1554_1600"}

    def on_message(self, ws, response):
        """
        """
        if 'chatpush' in response:
            response = MessageResponse(response[2:])
            print(response)

    def on_open(self, ws):
        print('connection established')

        count = 420
        for key in list(self.topwar_channels.keys()):
            channel_id = self.topwar_channels[key]
            print(f'Joined {key.upper()}')
            ws.send(f'{count}["join","{channel_id}"]')
            count += 1

        print(f'Binded UUID: {self.uuid}')
        ws.send(f'{count}["bind","{self.uuid}"]')

        def start_heartbeat():
            time.sleep(ping_interval)
            ws.send('2')
            start_heartbeat()

        threading.Thread(target=start_heartbeat).start()

    def on_close(self, ws):
        print('disconnected')

    def connect_websocket(self):
        self.ws = websocket.WebSocketApp(self.web_socket,
                                         on_message=self.on_message,
                                         on_close=self.on_close,
                                         on_open=self.on_open)
        self.ws.run_forever()

    def run(self):
        self.connect_websocket()


TopWarMessage().run()



# -*- coding: utf-8 -*-

import logging
import os
import platform
import subprocess

from notetool.tool.log import logger


def check_phone():
    """
    查看连接的手机
    """
    stdout, stderr = execute_cmd("adb devices")

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    logger.info(stdout)
    return stdout, stderr


def execute_cmd(cmd):
    execute = subprocess.Popen(
        str(cmd), stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
    stdout, stderr = execute.communicate()
    return stdout, stderr


def copy_file_to_local(from_path='/sdcard/screenshots.png', to_path="screenshots.png"):
    """
    从手机复制文件到电脑
    :param from_path: 手机文件位置
    :param to_path: 本地文件位置
    :return:
    """
    cmd = r"adb pull " + from_path + " " + to_path
    stdout, stderr = execute_cmd(cmd)
    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    logger.info("copy file from " + from_path + " to " + to_path)
    return stdout, stderr


def screenshot_to_phone(file_path='/sdcard/screenshots.png'):
    """
    在手机上截图
    :param file_path: 文件保存位置
    :return:
    """
    cmd = r"adb shell /system/bin/screencap -p " + file_path
    stdout, stderr = execute_cmd(cmd)

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    logger.info("screenshots and save to phone " + file_path)
    return stdout, stderr


def screenshot_to_local(file_path='screenshots.png'):
    """
    在手机上截图并复制到本地
    :param file_path:
    :return:
    """
    screenshot_to_phone()
    copy_file_to_local(to_path=file_path)


def swipe(x1, y1, x2, y2):
    cmd = r"adb shell input swipe " + \
        str(x1) + " " + str(y1) + " " + str(x2) + " " + str(y2)

    stdout, stderr = execute_cmd(cmd)

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    print(stdout)
    print(stderr)
    logging.info("swipe position " + str(x1) + " " +
                 str(y1) + " " + str(x2) + " " + str(y2))
    return stdout, stderr


def click(axis_x, axis_y):
    cmd = r"adb shell input tap " + str(axis_x) + " " + str(axis_y)

    stdout, stderr = execute_cmd(cmd)

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    print(stdout)
    print(stderr)
    logging.info("click position " + str(axis_x) + " " + str(axis_y))
    return stdout, stderr


def connect_ip(ip, port="5555"):
    cmd = r"adb connect " + ip + ":" + port
    stdout, stderr = execute_cmd(cmd)

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    logging.info("connected to " + ip + ":" + port)
    return stdout, stderr


def disconnect_ip(ip, port="5555"):
    cmd = r"adb disconnect " + ip + ":" + port
    stdout, stderr = execute_cmd(cmd)

    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")

    logging.info("disconnected to " + ip + ":" + port)
    return stdout, stderr


class AutoADB():
    def __init__(self):
        try:
            adb_path = 'adb'
            subprocess.Popen([adb_path], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
            self.adb_path = adb_path
        except OSError:
            if platform.system() == 'Windows':
                adb_path = os.path.join('Tools', "adb", 'adb.exe')
                try:
                    subprocess.Popen(
                        [adb_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    self.adb_path = adb_path
                except OSError:
                    pass
            else:
                try:
                    subprocess.Popen(
                        [adb_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                except OSError:
                    pass
            print('请安装 ADB 及驱动并配置环境变量')
            print('具体链接: https://github.com/wangshub/wechat_jump_game/wiki')
            exit(1)

    def get_screen(self):
        process = os.popen(self.adb_path + ' shell wm size')
        output = process.read()
        return output

    def run(self, raw_command):
        command = '{} {}'.format(self.adb_path, raw_command)
        process = os.popen(command)
        output = process.read()
        return output

    def test_device(self):
        print('检查设备是否连接...')
        command_list = [self.adb_path, 'devices']
        process = subprocess.Popen(
            command_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = process.communicate()
        if output[0].decode('utf8') == 'List of devices attached\n\n':
            print('未找到设备')
            print('adb 输出:')
            for each in output:
                print(each.decode('utf8'))
            exit(1)
        print('设备已连接')
        print('adb 输出:')
        for each in output:
            print(each.decode('utf8'))

    def test_density(self):
        process = os.popen(self.adb_path + ' shell wm density')
        output = process.read()
        return output

    def test_device_detail(self):
        process = os.popen(self.adb_path + ' shell getprop ro.product.device')
        output = process.read()
        return output

    def test_device_os(self):
        process = os.popen(
            self.adb_path + ' shell getprop ro.build.version.release')
        output = process.read()
        return output


"""
1： adb shell input keyevent keycode
   这条命令模拟Android手机按下了event_code对应的按键。
   event_code表如下：
0 -->  "KEYCODE_UNKNOWN"
1 -->  "KEYCODE_MENU"
2 -->  "KEYCODE_SOFT_RIGHT"
3 -->  "KEYCODE_HOME"
4 -->  "KEYCODE_BACK"
5 -->  "KEYCODE_CALL"
6 -->  "KEYCODE_ENDCALL"
7 -->  "KEYCODE_0"
8 -->  "KEYCODE_1"
9 -->  "KEYCODE_2"
10 -->  "KEYCODE_3"
11 -->  "KEYCODE_4"
12 -->  "KEYCODE_5"
13 -->  "KEYCODE_6"
14 -->  "KEYCODE_7"
15 -->  "KEYCODE_8"
16 -->  "KEYCODE_9"
17 -->  "KEYCODE_STAR"
18 -->  "KEYCODE_POUND"
19 -->  "KEYCODE_DPAD_UP"
20 -->  "KEYCODE_DPAD_DOWN"
21 -->  "KEYCODE_DPAD_LEFT"
22 -->  "KEYCODE_DPAD_RIGHT"
23 -->  "KEYCODE_DPAD_CENTER"
24 -->  "KEYCODE_VOLUME_UP"
25 -->  "KEYCODE_VOLUME_DOWN"
26 -->  "KEYCODE_POWER"
27 -->  "KEYCODE_CAMERA"
28 -->  "KEYCODE_CLEAR"
29 -->  "KEYCODE_A"
30 -->  "KEYCODE_B"
31 -->  "KEYCODE_C"
32 -->  "KEYCODE_D"
33 -->  "KEYCODE_E"
34 -->  "KEYCODE_F"
35 -->  "KEYCODE_G"
36 -->  "KEYCODE_H"
37 -->  "KEYCODE_I"
38 -->  "KEYCODE_J"
39 -->  "KEYCODE_K"
40 -->  "KEYCODE_L"
41 -->  "KEYCODE_M"
42 -->  "KEYCODE_N"
43 -->  "KEYCODE_O"
44 -->  "KEYCODE_P"
45 -->  "KEYCODE_Q"
46 -->  "KEYCODE_R"
47 -->  "KEYCODE_S"
48 -->  "KEYCODE_T"
49 -->  "KEYCODE_U"
50 -->  "KEYCODE_V"
51 -->  "KEYCODE_W"
52 -->  "KEYCODE_X"
53 -->  "KEYCODE_Y"
54 -->  "KEYCODE_Z"
55 -->  "KEYCODE_COMMA"
56 -->  "KEYCODE_PERIOD"
57 -->  "KEYCODE_ALT_LEFT"
58 -->  "KEYCODE_ALT_RIGHT"
59 -->  "KEYCODE_SHIFT_LEFT"
60 -->  "KEYCODE_SHIFT_RIGHT"
61 -->  "KEYCODE_TAB"
62 -->  "KEYCODE_SPACE"
63 -->  "KEYCODE_SYM"
64 -->  "KEYCODE_EXPLORER"
65 -->  "KEYCODE_ENVELOPE"
66 -->  "KEYCODE_ENTER"
67 -->  "KEYCODE_DEL"
68 -->  "KEYCODE_GRAVE"
69 -->  "KEYCODE_MINUS"
70 -->  "KEYCODE_EQUALS"
71 -->  "KEYCODE_LEFT_BRACKET"
72 -->  "KEYCODE_RIGHT_BRACKET"
73 -->  "KEYCODE_BACKSLASH"
74 -->  "KEYCODE_SEMICOLON"
75 -->  "KEYCODE_APOSTROPHE"
76 -->  "KEYCODE_SLASH"
77 -->  "KEYCODE_AT"
78 -->  "KEYCODE_NUM"
79 -->  "KEYCODE_HEADSETHOOK"
80 -->  "KEYCODE_FOCUS"
81 -->  "KEYCODE_PLUS"
82 -->  "KEYCODE_MENU"
83 -->  "KEYCODE_NOTIFICATION"
84 -->  "KEYCODE_SEARCH"
85 -->  "TAG_LAST_KEYCODE"
2：adb shell input tap
    这条命令模拟Android手机在屏幕坐标（X,Y）处进行了点击操作。
3：adb shell input swipe  
    这条命令模拟Android手机从屏幕坐标（X1,Y1）滑动到坐标（X2,Y2）的操作。
4、uiautomator dump   dump: creates an XML dump of current UI hierarchy 这个命令是用来成成当前界面的UI层次，并用XML格式进行展示 。这样就可以获取各个组件的位置了
注：如果PC要想同时控制多台Android手机，必须在adb 后面添加-s
例如：adb -s 13b6e4c4 shell input tap 400 400
表示对13b6e4c4这台Android手机进行在屏幕上（400,400）坐标位置进行模拟的点击事件。
"""
# check_phone()

# # screenshot_to_local()

# # click(109, 274)
# swipe(0, 500, 10, 800)



# -*- coding: utf-8 -*-
from abc import ABC
from collections import namedtuple
from copy import copy

import numpy as np
from notetool.tool.log import logger
from six.moves import map, range, zip

from notegame.games.nonogram.core.common import (BOX, SPACE, UNKNOWN, invert,
                                                 is_color_cell,
                                                 normalize_description)
from notegame.games.nonogram.core.renderer import Renderer
from notegame.games.nonogram.utils.iter import avg


class CellPosition(namedtuple('Cell', 'row_index column_index')):
    """2-D coordinates of a grid cell"""


class CellState(namedtuple('CellState', 'row_index column_index color')):
    """2-D coordinates of a grid cell with fixed color"""

    @property
    def position(self):
        """Get only coordinates (ignore the color)"""
        return CellPosition(self[0], self[1])

    @classmethod
    def from_position(cls, position, color):
        """Create a colored state from the CellPosition and color"""
        return cls(position[0], position[1], color)


def call_if_callable(func, *args, **kwargs):
    """Call the function with given parameters if it is callable"""
    if func and callable(func):
        return func(*args, **kwargs)

    return None


class RenderedMixin(object):
    """
    This mixin adds an ability to render the grid, i.e. a rectangular table of cells
    """

    def __init__(self, **kwargs):
        self.renderer = None
        self.set_renderer(**kwargs)

    def set_renderer(self, renderer=Renderer, **kwargs):
        """
        Allow to specify renderer even in the middle of the solving

        :type renderer: Renderer | type[Renderer]
        """

        if isinstance(renderer, type):
            self.renderer = renderer(self, **kwargs)
        elif isinstance(renderer, Renderer):
            self.renderer = renderer
            self.renderer.board_init(self)
        else:
            raise TypeError('Bad renderer: {!r}'.format(renderer))

    def draw(self, cells=None):
        """Draw a grid with the renderer"""
        self.renderer.draw(cells=cells)


class NonogramGrid(RenderedMixin):
    """
    Abstract nonogram grid with descriptions and cells
    """

    def __init__(self, columns, rows, cells=None, **renderer_params):
        """
        :param columns: iterable of vertical clues
        :param rows: iterable of horizontal clues
        """

        self.columns_descriptions = self.normalize(columns)
        self.rows_descriptions = self.normalize(rows)
        self.cells = cells or self.make_cells()
        self.validate()

        # setup the renderer after all the validations
        super(NonogramGrid, self).__init__(**renderer_params)

        # callbacks
        self.on_row_update = None
        self.on_column_update = None

    @property
    def init_cell_state(self):
        """Initial value of a single cell"""
        raise NotImplementedError()

    @property
    def height(self):
        """The height of the playing area"""
        return len(self.rows_descriptions)

    @property
    def width(self):
        """The width of the playing area"""
        return len(self.columns_descriptions)

    def make_cells(self):
        """Construct default cells set"""
        init_state = self.init_cell_state
        return [[init_state] * self.width for _ in range(self.height)]

    def normalize(self, clues):
        """
        Present clues in standard format
        """
        raise NotImplementedError()

    def validate(self):
        """
        Validate that the grid is valid:
        - all the descriptions of a row (column) can fit into width (height) of the grid
        - the vertical and horizontal descriptions define the same colors
        """
        self.validate_descriptions_size(self.columns_descriptions, self.height)
        self.validate_descriptions_size(self.rows_descriptions, self.width)
        self.validate_colors(self.columns_descriptions, self.rows_descriptions)

    @classmethod
    def validate_descriptions_size(cls, descriptions, max_size):
        """
        Validate that all the clues can fit into the grid
        """
        raise NotImplementedError()

    def validate_colors(self, vertical, horizontal):
        """
        Validate that the colors defined in vertical
        and horizontal descriptions are the same.
        """
        raise NotImplementedError()

    @classmethod
    def colors(cls):
        """
        All the colors that can appear on a grid

        :rtype: set
        """
        raise NotImplementedError()

    def unset_color(self, cell_state):
        """
        Drop the state from the list of possible states
        for a given cell
        :type cell_state: CellState
        """
        raise NotImplementedError()

    def set_color(self, cell_state):
        """
        Set the color of a cell with given coordinates
        :type cell_state: CellState
        """
        row_index, column_index, color = cell_state
        self.cells[row_index][column_index] = color

    @property
    def is_colored(self):
        """
        Whether the grid has an ability
        to store more than 2 colors (black and white only?)

        That is simpler than do `isinstance(board, ColoredBoard)` every time.
        """
        raise NotImplementedError()

    @property
    def has_blots(self):
        """
        Whether there are unknown (blotted) clues
        """
        return False

    def get_row(self, index):
        """Get the grid's row at given index"""
        return self.cells[index]

    def get_column(self, index):
        """Get the grid's column at given index"""
        return (row[index] for row in self.cells)

    # noinspection PyUnusedLocal
    def set_row(self, index, value):
        """Set the grid's row at given index"""
        self.cells[index] = list(value)

        self.row_updated(index)

    # noinspection PyUnusedLocal
    def set_column(self, index, value):
        """Set the grid's column at given index"""
        for row_index, item in enumerate(value):
            self.cells[row_index][index] = item

        self.column_updated(index)

    def __str__(self):
        return '{}({}x{})'.format(self.__class__.__name__, self.height, self.width)

    def row_updated(self, row_index):
        """Run each time the row gets partially solved"""
        call_if_callable(self.on_row_update, row_index, board=self)

    def column_updated(self, column_index):
        """Run each time the column gets partially solved"""
        call_if_callable(self.on_column_update, column_index, board=self)


class SolvableGrid(NonogramGrid, ABC):
    """
    Nonogram grid with various solution metrics defined
    """

    def __init__(self, columns, rows, cells=None, **kwargs):
        super(SolvableGrid, self).__init__(
            columns, rows, cells=cells, **kwargs)

        self._finished = False

        # custom callbacks
        self.on_solution_round_complete = None

    @property
    def solution_rate(self):
        """How much the board's cells are close to the full solution"""
        # if self.is_solved_full:
        #     return 1

        size = self.width
        return avg(self.line_solution_rate(row, size=size) for row in self.cells)

    @property
    def is_solved_full(self):
        """Whether no unsolved cells left"""
        raise NotImplementedError()

    @classmethod
    def line_solution_rate(cls, row, size=None):
        """How many cells in a given line are known to be solved"""
        raise NotImplementedError()

    def row_solution_rate(self, index):
        """How many cells in a horizontal row are known to be box or space"""
        return self.line_solution_rate(self.get_row(index), size=self.width)

    def column_solution_rate(self, index):
        """How many cells in a vertical column are known to be box or space"""
        return self.line_solution_rate(self.get_column(index), size=self.height)

    @classmethod
    def cell_solution_rate(cls, cell):
        """How much the cell's value is close to solved"""
        raise NotImplementedError()

    def is_cell_solved(self, position):
        """
        Decide whether the cell is completely solved
        :type position: CellPosition
        :rtype: bool
        """
        raise NotImplementedError()

    def cell_colors(self, position):
        """
        All the possible states that the cell at given position can be in.

        :type position: CellPosition
        :rtype: set
        """
        if not self.is_cell_solved(position):
            return self.colors()

        i, j = position
        return {self.cells[i][j]}

    def solution_round_completed(self):
        """
        Runs each time all the rows and the columns
        of the board gets partially solved (one solution round is complete)
        """
        call_if_callable(self.on_solution_round_complete, board=self)

    @property
    def is_finished(self):
        """Return whether the solving is finished"""
        return self._finished

    def set_finished(self, finished=True):
        """Set the solving status (used by renderers)"""
        self._finished = finished


class MultipleSolutionGrid(NonogramGrid, ABC):
    """
    Nonogram grid with additional support to store multiple states of a grid.
    Also substituting for the current cells is supported.
    """

    def __init__(self, columns, rows, cells=None, **renderer_params):
        super(MultipleSolutionGrid, self).__init__(
            columns, rows, cells=cells, **renderer_params)

        self.solutions = []

        # custom callbacks
        self.on_solution_found = None
        self.on_restored = None

    @classmethod
    def diff(cls, old_cells, new_cells, have_deletions=False):
        """
        Yield the coordinates of cells that was changed
        in the second set of cells compared to the first one.
        """
        assert len(old_cells) == len(new_cells)
        assert len(old_cells[0]) == len(new_cells[0])

        for i, row in enumerate(new_cells):
            for j, new_cell in enumerate(row):
                old_cell = old_cells[i][j]

                if have_deletions:
                    if new_cell != old_cell:
                        yield i, j

                else:
                    if is_color_cell(old_cell):
                        if new_cell < old_cell:
                            yield i, j
                        else:
                            assert new_cell == old_cell

                    elif new_cell != old_cell:
                        # '%s: %s --> %s' % ((i, j), old_cell, new_cell)
                        assert old_cell == UNKNOWN
                        yield i, j

    def changed(self, old_cells):
        """
        Yield the coordinates of cells that was changed
        compared to the given set of cells.
        """
        return self.diff(old_cells, self.cells)

    def make_snapshot(self):
        """Safely save the current state of a board"""
        # the values of the cells just shallow copied here
        # do not do deepcopy to prevent too heavy tuple's `deepcopy`
        return [list(row) for row in self.cells]

    def restore(self, snapshot):
        """Restore the previously saved state of a board"""
        self.cells = snapshot
        self.restored(snapshot)

    def restored(self, snapshot):
        """
        Run each time a grid's cells restored
        """
        call_if_callable(self.on_restored, snapshot)

    def _current_state_in_solutions(self):
        for i, sol in enumerate(self.solutions):
            diff = next(self.diff(sol, self.cells, have_deletions=True), None)
            if diff is None:
                logger.info('The solution is the same as the %d-th', i)
                if i > 2:
                    # faster to find repeated solutions that appear lately
                    logger.debug(
                        'Move found solution to the beginning of the list')
                    self.solutions.insert(0, self.solutions.pop(i))
                return True

            logger.info(
                'The solution differs from %d-th one: first differ cell: %s', i, diff)

        return False

    def add_solution(self, copy_=True):
        """Save one of the full solutions"""

        logger.info('Found one of the solutions!')

        if self._current_state_in_solutions():
            logger.info('Solution already exists')
            return

        if copy_:
            cells = self.make_snapshot()
        else:
            cells = self.cells

        self.solution_found(cells)
        self.solutions.append(cells)

    def solution_found(self, solution):
        """
        Run each time a new unique solution found
        """
        call_if_callable(self.on_solution_found, solution)

    def draw_solutions(self, only_logs=False):
        """Render the solutions"""
        if not self.solutions:
            return

        logger.info('Number of full unique solutions: %s', len(self.solutions))

        if not only_logs:
            for solution in self.solutions:
                self.draw(cells=solution)

        if len(self.solutions) == 1:
            return

        logger.info('Diff')
        for i, sol1 in enumerate(self.solutions):
            for j, sol2 in enumerate(self.solutions[i + 1:]):
                j = j + (i + 1)
                diff = list(self.diff(sol1, sol2, have_deletions=True))
                logger.info('%d vs %d: %d', i, j, len(diff))


class ReducibleGrid(SolvableGrid, MultipleSolutionGrid, ABC):
    """
    A nonogram grid which supports reducing,
    i.e. temporary trimming the solved rows and columns
    with appropriate descriptions changes.
    """

    def __init__(self, columns, rows, cells=None, **renderer_params):
        super(ReducibleGrid, self).__init__(
            columns, rows, cells=cells, **renderer_params)

        # save original descriptions to support reducing
        self.descriptions = (self.columns_descriptions, self.rows_descriptions)

        self.solved_columns = None
        self.solved_rows = None

    @classmethod
    def _space_value(cls):
        raise NotImplementedError()

    @classmethod
    def _reduce_orthogonal_description(cls, col_desc, cell_value, first_rows=False):
        raise NotImplementedError()

    @classmethod
    def _reduce_edge(cls, cells, straight_desc, orthogonal_desc,
                     line_solution_rate_func, first=True):
        # top, bottom
        solved_rows = []

        if first:
            rows_enum = list(enumerate(zip(cells, straight_desc)))
        else:
            rows_enum = reversed(list(enumerate(zip(cells, straight_desc))))

        for row_index, (row, row_desc) in rows_enum:
            if line_solution_rate_func(row_index) != 1:
                break

            logger.info('Reducing solved row (column) %i: %r',
                        row_index, row_desc)

            if first:
                # remove from the board description
                removed_desc = straight_desc.pop(0)

                # remove the cells itself
                cells = cells[1:]

                # save solved
                solved_rows.append(row)
            else:
                removed_desc = straight_desc.pop()
                cells = cells[:-1]
                solved_rows.insert(0, row)

            logger.info('Removed description %r', removed_desc)

            for col_index, (cell, col_desc) in enumerate(
                    zip(row, orthogonal_desc)):
                if not col_desc:  # skip empty description
                    continue

                if cell == cls._space_value():
                    continue

                logger.info('Reducing orthogonal description %i: %r',
                            col_index, col_desc)
                cls._reduce_orthogonal_description(
                    col_desc, cell, first_rows=first)

        return solved_rows, cells

    def reduce(self):
        """
        Cut out fully solved lines from the edges of the board, e.g.

           1 1 1
           1 1 1 1            1 1 1
        4  X X X X
        1  ? 0 ? 0   -->    1 ? 0 ?
        1  ? X ? 0          1 ? X ?
        """

        columns_descriptions = [list(col_desc)
                                for col_desc in self.columns_descriptions]
        rows_descriptions = [list(row_desc)
                             for row_desc in self.rows_descriptions]

        original_size = self.height, self.width

        # ====== ROWS ====== #
        cells = self.make_snapshot()
        first_solved_rows, cells = self._reduce_edge(
            cells, rows_descriptions, columns_descriptions,
            self.row_solution_rate, first=True)
        self.restore(cells)  # to correctly check line_solution_rate further

        last_solved_rows, cells = self._reduce_edge(
            cells, rows_descriptions, columns_descriptions,
            self.row_solution_rate, first=False)
        self.restore(cells)  # to correctly check line_solution_rate further

        self.columns_descriptions = self.normalize(columns_descriptions)
        self.rows_descriptions = self.normalize(rows_descriptions)

        # ====== COLS ====== #
        # transpose the matrix
        width = len(cells[0])
        cells = [list(self.get_column(col_index))
                 for col_index in range(width)]
        first_solved_columns, cells = self._reduce_edge(
            cells, columns_descriptions, rows_descriptions,
            self.column_solution_rate, first=True)

        # transpose it back
        height = len(cells[0])
        # to correctly check line_solution_rate further
        self.restore([[col[row_index] for col in cells]
                      for row_index in range(height)])

        last_solved_columns, cells = self._reduce_edge(
            cells, columns_descriptions, rows_descriptions,
            self.column_solution_rate, first=False)

        # transpose it back
        height = len(cells[0])
        # to correctly check line_solution_rate further
        self.restore([[col[row_index] for col in cells]
                      for row_index in range(height)])

        self.columns_descriptions = self.normalize(columns_descriptions)
        self.rows_descriptions = self.normalize(rows_descriptions)

        for sol_index, solution in enumerate(self.solutions):
            if first_solved_columns:
                solution = [row[len(first_solved_columns):]
                            for row in solution]
            if last_solved_columns:
                solution = [row[:-len(last_solved_columns)]
                            for row in solution]

            if first_solved_rows:
                solution = solution[len(first_solved_rows):]
            if last_solved_rows:
                solution = solution[:-len(last_solved_rows)]

            assert len(solution) == self.height
            assert len(solution[0]) == self.width

            self.solutions[sol_index] = solution

        reduced_size = self.height, self.width

        if original_size == reduced_size:
            logger.warning('The board size: %r', original_size)
        else:
            logger.warning('Reduced the board: %r --> %r',
                           original_size, reduced_size)

        self.solved_columns = (first_solved_columns, last_solved_columns)
        self.solved_rows = (first_solved_rows, last_solved_rows)
        return self.solved_columns, self.solved_rows

    @classmethod
    def restore_cells(cls, cells, edge_rows, edge_columns):
        """
        Return matrix by given center and the edges.
        Current implementation restores the columns first then the rows.
        """
        restore_rows = False
        if edge_rows:
            if edge_rows[0] or edge_rows[1]:
                restore_rows = True

        restore_cols = False
        if edge_columns:
            if edge_columns[0] or edge_columns[1]:
                restore_cols = True

        if not restore_rows and not restore_cols:
            return cells

        # do not touch original
        cells = [list(row) for row in cells]

        if edge_columns:
            first, last = edge_columns
            # insert one column at a time
            for col in reversed(first):
                assert len(col) == len(cells)
                cells = [[col_cell] + row for col_cell, row in zip(col, cells)]

            # append one column at a time
            for col in last:
                assert len(col) == len(cells)
                cells = [row + [col_cell] for col_cell, row in zip(col, cells)]

        if edge_rows:
            first, last = edge_rows
            # insert one row at a time
            for row in reversed(first):
                assert len(row) == len(cells[-1])
                cells.insert(0, row)

            # append one column at a time
            for row in last:
                assert len(row) == len(cells[0])
                cells.append(row)

        return cells

    def restore_reduced(self):
        """
        Restore the original size of the board if it was reduced before.
        Do it before rendering or yielding the final result.
        """

        current = self.cells
        reduced_size = self.height, self.width
        assert reduced_size == (len(current), len(current[0]))

        for sol_index, solution in enumerate(self.solutions):
            cells = self.restore_cells(
                solution, self.solved_rows, self.solved_columns)
            self.restore(cells)
            self.solutions[sol_index] = self.cells

        cells = self.restore_cells(
            current, self.solved_rows, self.solved_columns)
        self.restore(cells)

        self.columns_descriptions, self.rows_descriptions = self.descriptions

        original_size = self.height, self.width
        assert original_size == (len(self.cells), len(self.cells[0]))

        if original_size != reduced_size:
            logger.warning('Restored the board: %r --> %r',
                           reduced_size, original_size)


class BaseBoard(ReducibleGrid, ABC):
    """
    Base nonogram board with support of the following:
    - solving metrics
    - storing multiple solutions
    - reducing the solved lines
    """

    def __init__(self, columns, rows, cells=None, **renderer_params):
        super(BaseBoard, self).__init__(
            columns, rows, cells=cells, **renderer_params)

        # True =_column; False = row
        self.densities = {
            True: [self.line_density(True, index) for index in range(self.width)],
            False: [self.line_density(False, index) for index in range(self.height)],
        }

    @classmethod
    def desc_sum(cls, desc):
        """Minimal length that will be sufficient to store the given description"""
        raise NotImplementedError()

    def line_density(self, is_column, index):
        """
        The value in range [0..1] that shows how dense will be the solved line.
        The minimum density (0) is for the empty description.
        The maximum density is for the description that does not allow extra spaces between blocks.

        In general, the more this value the easier this line has to be solved.
        """
        if is_column:
            desc = self.columns_descriptions[index]
            full = self.height
        else:
            desc = self.rows_descriptions[index]
            full = self.width

        density = self.desc_sum(desc) / full

        assert 0 <= density <= 1
        return density

    def neighbours(self, position):
        """
        For the given cell yield
        the four possible neighbour cells.
        When the given cell is on a border,
        that number can reduce to three or two.
        :type position: CellPosition
        """
        row_index, column_index = position
        if row_index > 0:
            yield row_index - 1, column_index

        if row_index < self.height - 1:
            yield row_index + 1, column_index

        if column_index > 0:
            yield row_index, column_index - 1

        if column_index < self.width - 1:
            yield row_index, column_index + 1

    def unsolved_neighbours(self, position):
        """
        For the given cell yield the neighbour cells
        that are not completely solved yet.
        :type position: CellPosition
        """
        for pos in self.neighbours(position):
            pos = CellPosition(*pos)
            if not self.is_cell_solved(pos):
                yield pos


class NumpyBoard(BaseBoard, ABC):
    """
    The board that stores its state in a numpy array
    """

    def __init__(self, columns, rows, cells=None, **renderer_params):
        super(NumpyBoard, self).__init__(
            columns, rows, cells=cells, **renderer_params)
        self.restore(self.cells)

    def get_column(self, index):
        # self.cells.transpose([1, 0, 2])[index]
        return self.cells.T[index]

    def set_column(self, index, value):
        self.cells[:, index] = value
        self.column_updated(index)

    def make_snapshot(self):
        return copy(self.cells)

    def restore(self, snapshot):
        self.cells = np.array(snapshot)

    def _current_state_in_solutions(self):
        for solution in self.solutions:
            if np.array_equal(self.cells, solution):
                return True

        return False


class BlackBoard(BaseBoard):
    """
    Black-and-white nonogram board
    """

    @property
    def init_cell_state(self):
        return UNKNOWN

    def normalize(self, clues):
        return tuple(map(normalize_description, clues))

    @classmethod
    def validate_descriptions_size(cls, descriptions, max_size):
        for clue in descriptions:
            need_cells = sum(clue)
            if clue:
                # also need at least one space between every two blocks
                need_cells += len(clue) - 1

            logger.debug('Clue: %s; Need: %s; Available: %s.'% (clue, need_cells, max_size))
            if need_cells > max_size:
                raise ValueError('Cannot allocate clue {} in just {} cells'.format(
                    list(clue), max_size))

    def validate_colors(self, vertical, horizontal):
        boxes_in_columns = sum(sum(block) for block in vertical)
        boxes_in_rows = sum(sum(block) for block in horizontal)
        if boxes_in_rows != boxes_in_columns:
            raise ValueError('Number of boxes differs: {} (rows) and {} (columns)'.format(
                boxes_in_rows, boxes_in_columns))

    def is_cell_solved(self, position):
        i, j = position
        cell = self.cells[i][j]
        return cell != UNKNOWN

    @classmethod
    def colors(cls):
        return {BOX, SPACE}

    @property
    def is_colored(self):
        return False

    def unset_color(self, cell_state):
        row_index, column_index, bad_state = cell_state
        if self.cells[row_index][column_index] != UNKNOWN:
            raise ValueError('Cannot unset already set cell %s' %
                             ([row_index, column_index]))
        self.cells[row_index][column_index] = invert(bad_state)

    @property
    def is_solved_full(self):
        for row in self.cells:
            for cell in row:
                if cell == UNKNOWN:
                    return False
        return True

    @classmethod
    def line_solution_rate(cls, row, size=None):
        """How many cells in a given line are known to be box or space"""

        if size is None:
            size = len(row)

        return sum(1 for cell in row if cell != UNKNOWN) / size

    @classmethod
    def cell_solution_rate(cls, cell):
        """Whether the cell solved or not"""

        if cell == UNKNOWN:
            return 0
        return 1

    @classmethod
    def desc_sum(cls, desc):
        if not desc:
            return 0

        return sum(desc) + (len(desc) - 1)

    @classmethod
    def _space_value(cls):
        return SPACE

    @classmethod
    def _reduce_orthogonal_description(cls, col_desc, cell_value, first_rows=False):
        assert cell_value == BOX
        if first_rows:
            first_block = col_desc[0]
            if first_block == 1:
                col_desc.pop(0)
            else:
                col_desc[0] = first_block - 1
        else:
            last_block = col_desc[-1]
            if last_block == 1:
                col_desc.pop()
            else:
                col_desc[-1] = last_block - 1


class NumpyBlackBoard(BlackBoard, NumpyBoard):
    """Black-and-white board that uses numpy matrix to store the cells"""


def _solve_on_space_hints(board, hints):
    """
    Pseudo solving with spaces given
    """
    # assert len(hints) == len(board.rows_descriptions)
    for i, (spaces_hint, row) in enumerate(zip(hints, board.rows_descriptions)):
        assert len(spaces_hint) == len(row)
        cells = []
        for space_size, box_size in zip(spaces_hint, row):
            cells.extend([SPACE] * space_size)
            cells.extend([BOX] * box_size)

        # pad with spaces
        solution = cells + ([SPACE] * (board.width - len(cells)))
        board.cells[i] = solution


def make_board(*args, **kwargs)->NumpyBlackBoard:
    """Produce the black-and-white or colored nonogram"""

    if len(args) != 2:
        raise ValueError('Bad number of *args')

    return NumpyBlackBoard(*args, **kwargs)
